package com.sillycycle.bagleyd.life2d;

/*
 * @(#)Life2DCanvas.java
 *
 * Copyright 1993 - 2022  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Random;
import javax.swing.JFrame;

import com.sillycycle.bagleyd.util.Cell2D;
import com.sillycycle.bagleyd.util.CellInfo;
import com.sillycycle.bagleyd.util.ModArith;
import com.sillycycle.bagleyd.util.Signatures;

/**
 * The <code>Life2DCanvas</code> class does all the graphics for
 * <code>Life2DFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class Life2DCanvas extends Canvas {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");

	static final int MAX_STATES = 4;
	static final int MAX_AGE = 127;
	static final String DATAFILE = "data.life";

	Random generator;
	int startSpeed = Life2DInterface.DEFAULT_SPEED;
	int speed = Life2DInterface.DEFAULT_SPEED;
	int columns = Life2DInterface.DEFAULT_COLUMNS;
	int rows = Life2DInterface.DEFAULT_ROWS;
	int ncols = columns;
	int nrows = rows;
	int minCol = columns / 2 - 1;
	int maxCol = columns / 2 + 1;
	int minRow = rows / 2 - 1;
	int maxRow = rows / 2 + 1;
	boolean randomOrient = true;
	boolean grid = true;
	boolean wrap = true;
	boolean visible = true;
	boolean mono = false;
	boolean round = false;
	boolean findStable = false;
	boolean unboundDetection = false;
	boolean edgeDetection = false;
	boolean focus = true;
	boolean mouseDown = false;
	boolean edge = false;
	boolean paused = true;
	boolean generatingDB = true;
	boolean scroll = false;
	int neighbors =
		Life2DRules.indexToNeighbor[Life2DInterface.DEFAULT_NEIGHBOR_INDEX];
	int birth = 0;
	int survival = 0;
	int excludeCount = Life2DInterface.DEFAULT_EXCLUDE_COUNT;
	int searchCount = Life2DInterface.DEFAULT_SEARCH_COUNT;
	int excludePeriod = Life2DInterface.DEFAULT_EXCLUDE_PERIOD;
	int searchLength = Life2DInterface.DEFAULT_SEARCH_LENGTH;
	int fillPercent = Life2DInterface.DEFAULT_FILL_PERCENT;
	int blobPercent = Life2DInterface.DEFAULT_BLOB_PERCENT;
	int stirs = Life2DInterface.DEFAULT_STIRS;
	Color initialCellColor = Color.red;
	Color gridColor = Color.gray;
	Color background = Color.black;
	Cell2D cellPl;
	transient Capturer capturer;
	JFrame frame = null;
	static final boolean DEBUG = false;
	String name = "newbie";
	Life2DMatrix field = null;
	transient Life2DSoup soup = null;
	int gen = 0, genOffset = 0;
	int nCells = 0, oldnCells = 0, cCells = 0, group = 0;
	int states = 2;
	String stringSave = null;
	transient Life2DRules param = null;
	String signature = null;
	String stamp = null;
	transient Signatures signatures = null;
	transient Signatures findSignatures = null;
	int[] findNCells = null;
	int[] findPeriod = null;
	int[] findExcludePeriod = null;
	int checkFind = -1;
	StringBuilder initSB = new StringBuilder();
	//static final char endOfPatternRLE = '!';
	//static final char comment = '#';

	private synchronized void pulse(boolean draw, boolean stirring) {
		cCells = 0;
		/* 1. */
		/* for each live cell in linked list
		 *   create a dead list if toggle 0 and field array 0
		 * set toggle in field */
		if (param.getBirth(0)) {
			for (int position = 0; position < getColumns() * getRows(); position++) {
				if (field.getCell(position) == 0) {
					field.setDeadCell(position);
				}
			}
		} else {
			// causing ConcurrentModificationException
			//for (Iterator<CellInfo> iterator = field.list.iterator();
			//		iterator.hasNext();) {
			//	CellInfo cellInfo = iterator.next();
			//}
			for (int i = 0; i < field.list.size(); i++) {
				CellInfo cellInfo = field.list.get(i);
				int position = cellInfo.getPosition();
				int col = getColumn(position);
				int row = getRow(position);
				for (int neighbor = 0; neighbor < Math.abs(neighbors); neighbor++)
					field.setDeadCell(positionOfNeighbor(neighbor, col, row));
			}
		}
		/* 2. */
		/* for each dead cell: reset toggles back to 0;
		 *   if comes alive keep, else remove */
		for (Iterator<CellInfo> iterator = field.deadList.iterator();
				iterator.hasNext();) {
			CellInfo cellInfo = iterator.next();
			int deadPosition = cellInfo.getPosition();
			int col = getColumn(deadPosition);
			int row = getRow(deadPosition);
			int fertility = getNeighbors(col, row); // new group
			byte state;

			if (stirring && stirs > gen) {
				state = (byte) ((param.getBirthStir(fertility,
					group)) ? (states - 1) : 0);
			} else {
				state = (byte) ((param.getBirth(fertility,
					group)) ? (states - 1) : 0);
			}
			field.resetCellBit(deadPosition);
			if (state == 0) {
				iterator.remove();
			}
		}
		/* 3. */
		/* for each live cell set toggle if dies */
		for (CellInfo cellInfo : field.list) {
			int position = cellInfo.getPosition();
			int col = getColumn(position);
			int row = getRow(position);
			int value = getNeighbors(col, row); // new group
			byte nextState;
			int extra = 0;

			if (states > 2) {
				byte state;
				extra = 1;
				state = cellInfo.getCell();
				if (state == 3) {
					if (stirring && stirs > gen)
						nextState = (byte) ((param.getSurvivalStir(value +
							extra, group)) ? (states - 1) : 0);
					else
						nextState = (byte) ((param.getSurvival(value +
							extra, group)) ? (states - 1) : 0);
					if (nextState != state)
						cellInfo.setTemp((byte) (nextState + 1)); // come back later and set cell to 0
				} else { //if (state == 1 or 2)
					byte nextState0, nextState1;

					if (stirring && stirs > gen)
						nextState0 = (byte) ((param.getBirthStir(value + 1, group)) ? 1 : 0);
					else
						nextState0 = (byte) ((param.getBirth(value + 1, group)) ? 1 : 0);
					if (stirring && stirs > gen)
						nextState1 = (byte) ((param.getSurvivalStir(value, group)) ? 1 : 0);
					else
						nextState1 = (byte) ((param.getSurvival(value, group)) ? 1 : 0);
					if (state == 1)
						nextState = (byte) (((nextState0 << 1) & 0xff) + nextState1);
					else
						nextState = (byte) (((nextState1 << 1) & 0xff) + nextState0);
					if (nextState != state)
						cellInfo.setTemp((byte) (nextState + 1)); // come back later and set cell to 0
				}
			} else {
				if (stirring && stirs > gen)
					nextState = (byte) ((param.getSurvivalStir(value +
						extra, group)) ? (states - 1) : 0);
				else
					nextState = (byte) ((param.getSurvival(value +
						extra, group)) ? (states - 1) : 0);
				if (nextState == 0)
					cellInfo.setTemp((byte) 1); // come back later and set cell to 0
			}
		}
		/* 4. */
		/* for each live cell update according to toggle
		 *   if dies erase cell remove.
		 *   if lives update age, draw if necessary. */
		for (Iterator<CellInfo> iterator = field.list.iterator();
				iterator.hasNext();) {
			CellInfo cellInfo = iterator.next();
			int position = cellInfo.getPosition();
			byte temp = cellInfo.getTemp();
			switch (temp) {
			case 0:
				if (states == 2) {
					byte age = cellInfo.getCell();
					if (!mono && draw && age != MAX_AGE) {
						age = (byte) (age + 1);
						cellInfo.setCell(age);
						drawCell(cellInfo);
						field.setArrayCell(position, age);
					}
					//cellInfo.setCell(field.getCell(position));
				}
				break;
			case 1:
				if (draw) {
					eraseCell(cellInfo);
				}
				field.setArrayCell(position, (byte) 0);
				nCells--;
				cCells++;
				iterator.remove();
				break;
			default:
				temp = (byte) (temp - 1);
				cellInfo.setTemp((byte) 0);
				cellInfo.setCell(temp);
				field.setArrayCell(position, temp);
				if (draw) {
					drawCell(cellInfo);
				}
				cCells++;
			}
		}
		/* 5. */
		/* for each dead cell update to live cell. */
		for (Iterator<CellInfo> iterator = field.deadList.iterator();
				iterator.hasNext();) {
			CellInfo cellInfo = iterator.next();
			int deadPosition = cellInfo.getPosition();
			int col = getColumn(deadPosition);
			int row = getRow(deadPosition);
			byte nextState = (byte) (states - 1);
			cellInfo.setCell(nextState);
			if (draw)
				drawCell(cellInfo);
			field.setCell(deadPosition, nextState);
			checkAliveBounds(col, row);
			nCells++;
			cCells++;
			iterator.remove();
		}
		if (nCells > 0) {
			gen++;
			if (gen - stirs == 1 && unboundDetection)
				saveInitLife();
		} else {
			paused = true;
		}
	}

	void lifePulse(boolean draw, boolean fast, boolean stirring) {
		pulse(draw, stirring);
		if (draw && (!fast || (gen & 63) == 0)) {
			((Life2DFrame) frame).updateGen(gen);
			((Life2DFrame) frame).updateCells(nCells);
		}
	}

	public Life2DCanvas() {
		generator = new Random(System.nanoTime());
		capturer = new Capturer();
		soup = new Life2DSoup(this);
	}

	void inits() {
		resizeField();
	}

	void ruleChange() {
		if ((Life2DFrame) frame != null) {
			param = new Life2DRules(neighbors,
				((Life2DFrame) frame).getBirth(),
				((Life2DFrame) frame).getSurvival(),
				Life2DFrame.getBirthGroup(),
				Life2DFrame.getSurvivalGroup());
			String ruleName = param.getRuleName();

			if (ruleName == null)
				((Life2DFrame) frame).updateRule(Integer.toString(neighbors));
			else
				((Life2DFrame) frame).updateRule(neighbors + ruleName);
			if (DEBUG)
				System.out.println("ruleChange: " + ruleName);
		}
		signatures = null;
	}

	public int getStates() {
		return this.states;
	}

	public int getNeighborIndex() {
		return Life2DRules.neighborToIndex(neighbors);
	}

	Integer getRuleIndex(int neighborIndex) {
		String ruleName = param.getRuleName();

		if (ruleName == null || ruleName.isEmpty())
			return Life2DRules.defaultRuleIndex(neighborIndex);
		ruleName = Life2DRules.parseRuleCondense(param.getRuleName());
		if (ruleName == null) {
			return null;
		}
		if (DEBUG)
			System.out.println("getRuleIndex: " + ruleName);
		return param.getRuleIndex(neighborIndex, ruleName);
	}

	static boolean polygonChange(int oldNeighborIndex, int newNeighborIndex) {
		if (oldNeighborIndex != newNeighborIndex) {
			int oldPolygon = Life2DRules.neighborIndexToPolygon[oldNeighborIndex];
			int newPolygon = Life2DRules.neighborIndexToPolygon[newNeighborIndex];

			return (oldPolygon != newPolygon);
		}
		return false;
	}

	int getPolygon() {
		return Life2DRules.neighborToPolygon(neighbors);
	}

	void setNeighbors(int neighborIndex) {
		neighbors = Life2DRules.indexToNeighbor[neighborIndex];
	}

	int getRealColumns(int cols) {
		int polygon = getPolygon();
		int realColumns = cols;

		if (polygon == 3 || polygon == 6) {
			if ((realColumns & 1) == 1)
				realColumns--;
		} else if (polygon == 5) {
			realColumns -= realColumns % 4;
		}
		return realColumns;
	}

	int getRowsFromColumns(int cols) {
		int polygon = getPolygon();
		int calculatedRows = 16 * cols / 25;

		if ((polygon == 3 || polygon == 5 || polygon == 6) &&
				((calculatedRows & 1) == 1)) {
			calculatedRows--;
		}
		return calculatedRows;
	}

	public int getColumns() {
		return this.ncols;
	}

	public int getRows() {
		return this.nrows;
	}

	public void speedUp() {
		speed += 10;
		if (speed > Life2DInterface.MAX_SPEED)
			speed = Life2DInterface.MAX_SPEED;
		((Life2DFrame) frame).setDelay(Life2DInterface.MAX_SPEED +
			Life2DInterface.MIN_SPEED - 1 - speed);
	}

	public void slowDown() {
		speed -= 10;
		if (speed < Life2DInterface.MIN_SPEED)
			speed = Life2DInterface.MIN_SPEED;
		((Life2DFrame) frame).setDelay(Life2DInterface.MAX_SPEED +
			Life2DInterface.MIN_SPEED - 1 - speed);
	}

	void resizeField() {
		int newWidth = getSize().width;
		int newHeight = getSize().height;
		int polygon = getPolygon();

		capturer.clearCapture();
		int tempCols = Life2DFrame.getColumns();
		ncols = getRealColumns(tempCols);
		nrows = getRowsFromColumns(tempCols);
		if (newWidth == 0)
			newWidth = Life2DInterface.DEFAULT_CELLSIZE *
				ncols + 4;
		if (newHeight == 0)
			newHeight = Life2DInterface.DEFAULT_CELLSIZE *
				nrows + 4;
		cellPl = new Cell2D(polygon, ncols, nrows, newWidth, newHeight,
			(Life2DFrame) frame != null && ((Life2DFrame)frame).getScroll());
//System.out.println("Resize " + neighbors);
//System.out.println("ncols " + ncols + ", " + nrows);
//System.out.println("cols " + columns + ", " + rows);
//System.out.println(param.getCommonName());
		if (field == null) {
			columns = ncols;
			rows = nrows;
			field = new Life2DMatrix(ncols, nrows);
			gen = 0;
			genOffset = 0;
			minCol = ncols / 2 - 1;
			maxCol = ncols / 2 + 1;
			minRow = nrows / 2 - 1;
			maxRow = nrows / 2 + 1;
			nCells = 0;
			if ((Life2DFrame) frame != null) {
				((Life2DFrame) frame).initPopup();
			}
		} else if (columns != ncols || rows != nrows) {
			/*if (polygon == 3) {
				if (((minCol + minRow) & 1) == 1) {
					if ((minRow & 1) == 1) {
						minRow--;
					} else {
						minCol--;
					}
				}
			} else if (polygon == 5) {
				int mod = (minCol + 2 * minRow) % 4;

				if ((mod / 2) == 1) {
					if ((minRow & 1) == 1) {
						minRow--;
					} else {
						minCol -= 2 ;
					}
				}
				minCol -= (mod & 1);
			} else if (polygon == 6 && (minRow & 1) == 1) {
				minRow--;
			}*/
			recalibrateBounds(columns, rows);

			int fieldX = maxCol - minCol + 1;
			int fieldY = maxRow - minRow + 1;
			Life2DMatrix cage = new Life2DMatrix(fieldX, fieldY);
			for (CellInfo cellInfo : field.list) {
				int oldP = cellInfo.getPosition();
				int i = getColumn(oldP, columns);
				int j = getRow(oldP, columns);
				byte value = field.getCell(oldP);

				if (value != 0) {
					int p = getPosition(i - minCol, j - minRow,
						fieldX);
					cage.setArrayCell(p, value);
				}
			}
			field.clearLiveCells();
			columns = ncols;
			rows = nrows;
			field = new Life2DMatrix(ncols, nrows);
			gen = 0;
			genOffset = 0;
			minCol = ncols / 2 - 1;
			maxCol = ncols / 2 + 1;
			minRow = nrows / 2 - 1;
			maxRow = nrows / 2 + 1;
			nCells = 0;

			int offsetX = minCol + 1 - fieldX / 2;
			int offsetY = minRow + 1 - fieldY / 2;

			if (polygon == 3) {
				if (((offsetX + offsetY) & 1) == 1) {
					if ((offsetY & 1) == 1) {
						offsetY--;
					} else {
						offsetX--;
					}
				}
			} else if (polygon == 5) {
				int mod = (offsetX + 2 * offsetY) % 4;

				if ((mod / 2) == 1) {
					if ((offsetY & 1) == 1) {
						offsetY--;
					} else {
						offsetX -= 2 ;
					}
				}
				offsetX -= (mod & 1);
			} else if (polygon == 6 && (offsetY & 1) == 1) {
				offsetY--;
			}
			for (int j = 0; j < fieldY; j++)
			if ((j + offsetY >= 0) && (j + offsetY <= nrows - 1))
			for (int i = 0; i < fieldX; i++)
			if ((i + offsetX >= 0) && (i + offsetX <= ncols - 1)) {
				int oldP = getPosition(i, j, fieldX);
				byte value = cage.getCell(oldP);

				if (value != 0) {
					int p = getPosition(i + offsetX, j + offsetY,
						ncols);
					field.setCell(p, value);
					nCells++;
				}
			}
			if ((Life2DFrame) frame != null) {
				((Life2DFrame) frame).updateGen(gen);
				((Life2DFrame) frame).updateCells(nCells);
				((Life2DFrame) frame).initPopup();
			}
		}
	}

	void xorCell(Graphics g, CellInfo info, byte state, int k) {
		int col = getColumn(info);
		int row = getRow(info);
		Color color;

		if (state == 0) {
			color = background;
		} else {
			if (mono) {
				color = Color.white;
			} else if (states == 2) {
				color = colorRainbow(state, 16);
			} else {
				color = colorRainbow((state - 1) * 16, 16);
			}
		}
		if ((k & 1) == 0) // Simulate Xor
			color = new Color(255 - color.getRed(),
				255 - color.getGreen(),
				255 - color.getBlue());
		cellPl.drawCell(g, round, color,
			(grid) ? gridColor : background, col, row);
	}

	// Should make some attempt at minimizing size if on a border...
	void printField(StringBuilder sb) {
		String space = " ";
		String newline = NL;
		String pound = "#";
		int polygon = getPolygon();
		boolean firstPoint = true;
		int i, j, p;
		int startX = 0, startY = 0;
		int endX = 0, endY = 0;

		sb.append(pound).append("Dim").append(space)
			.append(2).append(newline);
		sb.append(pound).append("Neighbors")
			.append(space).append(neighbors).append(newline);
		sb.append(pound).append("Rule").append(space)
			.append("B").append(Life2DRules.condense(Life2DRules.getBirthString(param.getRuleName())))
			.append("S").append(Life2DRules.condense(Life2DRules.getSurvivalString(param.getRuleName()))).append(newline);
		sb.append(pound).append("Cells").append(space)
			.append(nCells).append(newline);
		sb.append(pound).append("Name").append(space)
			.append(name).append(newline);
		/*sb.append(pound).append("Sig").append(space)
			.append("An A0 <> D1 Dn").append(newline);*/
		for (j = 0; j < nrows; j++)
		for (i = 0; i < ncols; i++) {
				p = getPosition(i, j);
				if (field.getCell(p) != 0) {
					if (firstPoint) {
						startX = i;
						startY = j;
						endX = i;
						endY = j;
						firstPoint = false;
					} else {
						if (i < startX)
							startX = i;
						if (j < startY)
							startY = j;
						if (i > endX)
							endX = i;
						if (j > endY)
							endY = j;
					}
				}
		}
		sb.append(pound + "Width" + space + (endX - startX + 1) + newline);
		sb.append(pound + "Height" + space + (endY - startY + 1) + newline);
		sb.append(pound + "P" + space + (startX - ncols / 2) + space +
			(startY - nrows / 2) + newline);
		//sb.append(pound + "P" + space + 0 + space + 0 + newline);
		switch (polygon) {
		case 3:
			if (((startX + startY) & 1) == 1) {
				startX--;
			}
			break;
		case 5:
			startX -= (startX + 2 * startY) % 4;
			break;
		case 6:
			int append = (endY - startY + 1) / 2;
			int low = endX + append - startX;
			int high = 0;
			int offset = 0;

			for (j = startY; j <= endY; j++) {
				for (i = startX + offset; i <= endX + append; i++) {
					p = getPosition(i, j);
					if (field.getCell(p) != 0) {
						if (low > i - startX - offset)
							low = i - startX - offset;
						if (high < i - startX - offset )
							high = i - startX - offset;
					}
				}
				if ((j & 1) == 1) {
					offset--;
					append--;
				}
			}
			append = high;
			offset = low;
			for (j = startY; j <= endY; j++) {
				for (int k = j; k < endY; k++) {
					sb.append(space);
				}
				for (i = startX + offset; i <= startX + append; i++) {
					if (i > startX + offset)
						sb.append(space);
					p = getPosition(i, j);
					if (field.getCell(p) == 0) {
						sb.append(Life2DInterface.DEFAULT_DEAD_CHAR);
					} else if (states == 2) {
						sb.append('O');
					} else {
						sb.append(field.getCell(p));
					}
				}
				if ((j & 1) == 1) {
					offset--;
					append--;
				}
				sb.append(newline);
			}
			return;
		default:
		}
		for (j = startY; j <= endY; j++) {
		for (i = startX; i <= endX; i++) {
				p = getPosition(i, j);
				if ((field.getCell(p)) == 0) {
					sb.append(Life2DInterface.DEFAULT_DEAD_CHAR);
				} else if (states == 2) {
					switch (polygon) {
					case 3:
						if (((i + j) & 1) == 1)
							sb.append('V');
						else
							sb.append('A');
						break;
					case 5:
						int mod = (i + 2 * j) % 4;

						//AVJD graphical letters to convey
						// up, down, left, right.  Since also used
						// in cookie output, less than/greater than
						// symbols did not work well.  Coincidence
						// that the output looks like my name.  :)
						switch (mod) {
						case 0:
							sb.append('A');
							break;
						case 1:
							sb.append('V');
							break;
						case 2:
							sb.append('J');
							break;
						case 3:
							sb.append('D');
							break;
						default:
							System.out.println("wrong mod " + mod);
						}
						break;
					default:
						sb.append(Life2DInterface.DEFAULT_ALIVE_CHAR);
					}
				} else {
					sb.append(field.getCell(p));
				}
		}
		sb.append(newline);
		}
	}

	void clearField(boolean draw) {
		field.clearLiveCells();
		if (draw)
			eraseFrame();
		gen = 0;
		genOffset = 0;
		param.setStirs();
		edge = false;
		minCol = ncols / 2 - 1;
		maxCol = ncols / 2 + 1;
		minRow = nrows / 2 - 1;
		maxRow = nrows / 2 + 1;
		nCells = 0;
		if (draw && (Life2DFrame) frame != null) {
			Graphics g = this.getGraphics();

			((Life2DFrame) frame).updateGen(gen);
			((Life2DFrame) frame).updateCells(nCells);
			if (!scroll)
				drawFrame(g, true);
			if (grid)
				drawGrid();
		}
	}

	void showMessage(String msg) {
		((Life2DFrame) frame).showMessage(msg);
	}

	/*void setCell(int col, int row) {
		if (col >= 0 && col < ncols &&
				row >= 0 && row < nrows) {
			setCell(col, row, (byte) 1, visible);
		}
	}*/

	void setCell(int col, int row, int state) {
		if (col >= 0 && col < ncols &&
				row >= 0 && row < nrows) {
			setCell(col, row, (byte) state, visible);
		}
	}

	public String getGridStamp() {
		return findGridStamp(new Rectangle(minCol, minRow,
			maxCol - minCol + 1,
			maxRow - minRow + 1));
	}

	/** Counts the number of live cells and adds the coordinates.
	 * Not as pretty as a signature but usually catches cases
	 * where life form just is reoriented.
	 * @param r
	 * @return grid stamp string
	 */
	public String findGridStamp(Rectangle r) {
		int p;
		int polygon = getPolygon();
		Point linear = new Point(0, 0);
		Point square = new Point(0, 0);

		for (int j = r.y; j < r.y + r.height; j++)
		for (int i = r.x; i < r.x + r.width; i++) {
			p = getPosition(i, j);
			if (field.getCell(p) != 0) {
				int col = 0, row = 0;
			switch (polygon) {
			case 3:
				col = getColumn(p) - 2 * (ncols / 4) - 1;
				row = getRow(p) - 2 * (nrows / 4);
				break;
			case 4:
			case 5:
				col = getColumn(p) - (ncols / 2);
				row = getRow(p) - (nrows / 2);
				break;
			case 6:
				col = getColumn(p) - 2 * (ncols / 4);
				row = getRow(p) - 2 * (nrows / 4) - 1;
				if (row < 0)
					col += (((nrows / 2) & 1) == 1) ?
						(row - 1) / 2 : row / 2;
				else
					col += (((nrows / 2) & 1) == 1) ?
						row / 2 : (row + 1) / 2;
				break;
			default:
			}
			linear.x += col;
			linear.y += row;
			square.x += col * col;
			square.y += row * row;
				}
		}
		return "x" + linear.x + "y" + linear.y +
			"X" + square.x + "Y" + square.y;
	}

	public String getGridSignature() {
		return findGridSignature(new Rectangle(minCol, minRow,
			maxCol - minCol + 1,
			maxRow - minRow + 1));
	}

	/** Counts the number of live and dead cells with a certain
	 * number of live neighbors and then passes these "bucket"
	 * counts to another function which returns a formatted
	 * signature code for all cells currently in the grid.
	 * @param r
	 * @return grid signature string
	 */
	public String findGridSignature(Rectangle r) {
		int neighborCount, p;
		LifeSig sigCount = new LifeSig(Life2DRules.MAX_NEIGHBORS * states / 2);
		int minx = 0, maxx = ncols, miny = 0, maxy = nrows;
		int polygon = getPolygon();

		// need some corrections on non-square coordinates
		switch (polygon) {
		case 3:
			// possibly could be tighter
			if (r.x > minx)
				minx = r.x - 1;
			if (r.y > miny)
				miny = r.y - 1;
			if (r.x + r.width < ncols - 1)
				maxx = r.x + r.width + 1;
			if (r.y + r.height < nrows - 1)
				maxy = r.y + r.height + 1;
			break;
		case 4:
			minx = r.x;
			miny = r.y;
			maxx = r.x + r.width;
			maxy = r.y + r.height;
			break;
		case 5:
			// possibly could be tighter
			if (r.x > minx + 1)
				minx = r.x - 2;
			if (r.y > miny)
				miny = r.y - 1;
			if (r.x + r.width < ncols - 2)
				maxx = r.x + r.width + 2;
			if (r.y + r.height < nrows - 1)
				maxy = r.y + r.height + 1;
			break;
		default:
		}
		for (int j = miny; j < maxy; j++)
		for (int i = minx; i < maxx; i++) {
				int newI = (i + columns) % columns;
				int newJ = (j + rows) % rows;

				p = getPosition(newI, newJ);
				neighborCount = getNeighbors(newI, newJ);
				if (field.getCell(p) == 0) {
					sigCount.dead++;
					sigCount.deadCount[neighborCount]++;
				} else {
					sigCount.live++;
					sigCount.liveCount[neighborCount]++;
				}
		}
		return sigCount.signatureCode();
	}

	public boolean getSignatureList(boolean print,
			boolean db, boolean special, String species) {
		final int MAX_OSC = 160;
		final int MAX_SMOKE = 80;
		int osc = 0;
		int numCells;

		if (signatures == null) {
			signatures = new Signatures();
		}
		String[] gSignature = new String[MAX_OSC];
		gSignature[0] = getGridSignature();
		numCells = Signatures.signatureToNumberLiveCells(gSignature[0]);
		if (print && numCells != 0) {
			if (db) {
				if (species != null)
					System.out.print(species);
				System.out.print("[" + numCells + "]|");
			}
			System.out.println(gSignature[0]);
		}
		// This stops at first match.
		// This is not guaranteed to get all signatures
		// for example picture 2 clock like life forms
		// spinning opposite each other, they could have
		// same overall signature until they point at each
		// other.  Life forms can probably be entered in DB
		// in such a way that this is alleviated.

		if (special) {
			signatures.putSignature(gSignature[0]);
			return true;
		}
		while (osc < MAX_OSC - 1) {
			lifePulse(!db, true, false);
			String signatureN = getGridSignature();
			nCells = Signatures.signatureToNumberLiveCells(signatureN);
			gSignature[osc + 1] = signatureN;
			for (int smoke = 0; smoke < MAX_SMOKE; smoke++) {
				if (osc >= smoke && signatureN.equals(gSignature[smoke])) {
					for (int i = 0; i <= osc; i++) {
						signatures.putSignature(gSignature[i]);
					}
					return true;
				}
			}
			if (osc == MAX_OSC || nCells == 0 || "0 <> 0".equals(signatureN)) {
				if (species == null) // Screen junk
					return true;
				/*if (debug)*/
					System.out.println(species +
						": failed on osc " +
						osc + ", " + signatureN);
				return false;
			}
			if (print) {
				if (db) {
					if (species != null)
						System.out.print(species);
					System.out.print("[" + nCells + "]|");
				}
				System.out.println(signatureN);
			}
			osc++;
		}
		return false;
	}

	static int[] toIntArray(ArrayList<Integer> list) {
		int[] ret = new int[list.size()];
		int i = 0;
		for (Integer nonprim : list)
			ret[i++] = nonprim;
		return ret;
	}

	private void addSignatures(String string, boolean print) {
		String fileName = "sig/" + string + neighbors + "B" +
			Life2DRules.getBirthString(param.getRuleName()) + "S" +
			Life2DRules.getSurvivalString(param.getRuleName()) + ".sig";
		try (BufferedReader br = new BufferedReader(
				new FileReader(fileName))) {
			String lineData, sig;
			ArrayList<Integer> findNCellsList = null;
			ArrayList<Integer> findPeriodList = null;
			ArrayList<Integer> findExcludePeriodList = null;
			int mult;

			while ((lineData = br.readLine()) != null) {
				String[] lineToken;

				switch (string) {
				case "junk":
					sig = lineData;
					mult = signatures.containsSignature(sig);
					if (mult == 0) {
						signatures.putSignature(sig);
						if (print) {
							int numCells = Signatures.signatureToNumberLiveCells(sig);

							System.out.println(fileName + "[" + numCells + "]|" + sig);
						}
					} else {
						System.out.println(string + "(" + mult + "): " + sig);
					}
					break;
				case "find":
					lineToken = lineData.split("\\|", 2);
					if (lineToken.length > 1) {
						sig = lineToken[1];
						mult = signatures.containsSignature(sig);
						if (mult != 0)
							System.out.println(string + "(" + mult + "): " + sig);
						findSignatures.putSignature(sig);
						if (print) {
							int numCells = Signatures.signatureToNumberLiveCells(sig);

							System.out.println(lineToken[0] + "[" + numCells + "]|" + sig);
						}
					} else {
						String[] lineTokenParen = lineData.split("\\(p", 2);
						if (lineTokenParen.length > 1) {
							String[] lineTokenParen2 = lineTokenParen[1].split("\\)", 2);
							if (findPeriodList == null)
								findPeriodList = new ArrayList<>();
							findPeriodList.add(Integer.parseInt(lineTokenParen2[0].trim()));
						}
						String[] lineTokenBracket = lineData.split("\\[", 2);
						if (lineTokenBracket.length > 1) {
							String[] lineTokenBracket2 = lineTokenBracket[1].split("\\]", 2);
							if (findNCellsList == null)
								findNCellsList = new ArrayList<>();
							findNCellsList.add(Integer.parseInt(lineTokenBracket2[0].trim()));
						}
						String[] lineTokenOtherBracket = lineData.split("\\]", 2);
						if (lineTokenOtherBracket.length > 0) {
							if (findExcludePeriodList == null)
								findExcludePeriodList = new ArrayList<>();
							findExcludePeriodList.add(Integer.parseInt(lineTokenOtherBracket[1].trim()));
						}
					}
					break;
				default:
					lineToken = lineData.split("\\|", 2);
					if (lineToken.length > 1 &&
							signatures.containsSignature(lineToken[1]) == 0) {
						sig = lineToken[1];
						signatures.putSignature(sig);
						if (print) {
							int numCells = Signatures.signatureToNumberLiveCells(sig);

							System.out.println(lineToken[0] + "[" + numCells + "]|" + sig);
						}
					}
				}
			}
			if (findPeriodList != null) {
				findPeriod = toIntArray(findPeriodList);
			}
			if (findNCellsList != null) {
				findNCells = toIntArray(findNCellsList);
			}
			if (findExcludePeriodList != null) {
				findExcludePeriod = toIntArray(findExcludePeriodList);
			}
		} catch (IOException e) {
			//e.printStackTrace();
		}
	}

	public void generateDB(boolean print) {
		int[][] data;
		String patternName;
		int ni = getNeighborIndex(), ri;
		Integer ruleIndex = getRuleIndex(ni);
		int start = ((print) ? -1 : 0); // decide about on screen
		final int MAX_RETRY = 16; // arbitrary... but do not want it looping forever

		generatingDB = true;
		signatures = new Signatures();
		findSignatures = new Signatures();
		// For application read from junk list
		if (neighbors != Life2DRules.indexToNeighbor[ni])
			System.out.println("NeighborIndex: " + ni + ", Neighbors: " +
				neighbors);
		if (ruleIndex == null) {
			//System.out.println("For neighbor index " + ni +
			//	", ruleIndex is null");
			return;
		}
		ri = ruleIndex;
		if (!print)
			clearField(visible);
		for (Life2DForms.Category c : Life2DForms.Category.values()) {
			for (int fi = start; fi < Life2DForms.length(ni, ri, c); fi++) {
				for (int retry = 0; retry <= MAX_RETRY; retry++) {
					if (fi == -1) { // do what is on screen
						patternName = this.name;
					} else {
						data = Life2DForms.get(ni, ri, c, fi);
						patternName = Life2DForms.getName(ni, ri, c, fi);
						if (data == null || data[0].length == 0) {
							System.out.println(fi + " data for ruleIndex " + ri + " is null");
							return;
						}
						displayForm(data, false);
					}
					// This fails under heavy load
					if (getSignatureList(print, true, (fi != -1 && c == Life2DForms.Category.OTHER), patternName)) {
						clearField(false);
						break;
					}
					clearField(false);
					if (retry == MAX_RETRY && wrap)
						System.out.println("FIXME");
					/*try {
						Thread.sleep(100);
					} catch (InterruptedException e) {
					}*/
				}
			}
			start = 0;
		}
		addSignatures("found", print); // not needed if above works
		addSignatures("junk", print); // ignore these signatures
		if (print) {
			System.out.println("Number signatures: " +
				signatures.numberSignature());
			clearField(visible);
		}
		addSignatures("find", print);
		generatingDB = false;
	}

	private void setOrientCell(int x, int y,
			int startX, int startY,
			byte state, int orient, boolean draw) {
		int i = x, j = y, temp;
		int polygon = getPolygon();

		switch (polygon) {
		case 3:
			if (orient >= 6) {
				i = -i;
			}
			switch (orient % 6) {
			case 1:
				temp = j;
				j = ModArith.ceil2(temp + i);
				i = ModArith.ceil2(i - temp + 1) - temp;
				break;
			case 2:
				temp = j;
				j = ModArith.ceil2(temp - i);
				i = temp - ModArith.ceil2(1 - i - temp);
				break;
			case 3:
				i = -i;
				j = -j + 1;
				break;
			case 4:
				temp = j;
				j = 1 - ModArith.ceil2(temp + i);
				i = temp - ModArith.ceil2(i - temp + 1);
				break;
			case 5:
				temp = j;
				j = 1 - ModArith.ceil2(temp - i);
				i = ModArith.ceil2(1 - i - temp) - temp;
				break;
			case 0:
			default:
				break;
			}
			break;
		case 4:
			if (orient >= 4) {
				temp = i;
				i = j;
				j = temp;
			}
			if (orient % 4 >= 2) {
				i = -i;
			}
			if ((orient & 1) == 1) {
				j = -j;
			}
			break;
		case 5:
			int mod = ModArith.mod(i + 2 * j, 4);
			int ti = j * 2;
			int tj = ModArith.div(i, 2);

			if (mod == 1 || mod == 2)
				ti++;
			switch (orient % 4) {
			case 1:
				i = -ti - 1;
				j = tj;
				break;
			case 2:
				i = -i - 1;
				j = -j - 1;
				break;
			case 3:
				i = ti;
				j = -tj - 1;
				break;
			default:
				break;
			}
			if (orient >= 4) {
				i = -i + 1;
				mod = ModArith.mod(i + 2 * j, 4);
				if (mod == 0)
					i++;
				if (mod == 1)
					i--;
			}
			break;
		case 6:
			if (orient >= 6) {
				temp = i;
				i = j;
				j = temp;
			}
			switch (orient % 6) {
			case 1:
				temp = j;
				j = temp - i;
				i = temp;
				break;
			case 2:
				temp = -i;
				i = temp + j;
				j = temp;
				break;
			case 3:
				i = -i;
				j = -j;
				break;
			case 4:
				temp = -j;
				j = temp + i;
				i = temp;
				break;
			case 5:
				temp = i;
				i = temp - j;
				j = temp;
				break;
			case 0:
			default:
				break;
			}
			break;
		default:
		}
		if (polygon == 5) {
			temp = startX - (startX + startY * 2) % 4;
			i += temp;
		} else
			i += startX;
		if (polygon == 6) {
			if ((startX & 1) == 1 && j >= 0 || j > 0)
				i += (((startX & 1) == 1) ? -j / 2 :
					-(j - 1) / 2);
			else
				i += 1 + (((startX & 1) == 1) ? -(j + 1) / 2 :
					-j / 2);
		}
		j += startY;
		if (polygon == 3 &&
				(((startX + startY + 1) & 1) == 1)) {
			i++;
		}
		if (i >= 0 && i < ncols && j >= 0 && j < nrows)
			setCell(i, j, state, draw);
	}

	public void displayForm(int[][] data, boolean draw) {
		int i = 0, j, n;
		int startX, startY;
		int orient = 0;
		int polygon = getPolygon();

		startX = ncols / 2;
		startY = nrows / 2;
		if (polygon == 5) {
			startX -= (startX + startY * 2) % 4;
		} else if (polygon == 6 && (startY & 1) == 0) {
			startY -= 1;
		}
		if (randomOrient) {
			if (polygon == 3 || polygon == 6) {
				orient = generator.nextInt(12);
			} else if (polygon == 4 || polygon == 5) {
				orient = generator.nextInt(8);
			}
		}
		for (int s = 0; s < data.length && s < states - 1; s++) {
			n = 0;
			while (n < data[s].length) {
				if ((n & 1) == 1) {
					j = data[s][n];
					setOrientCell(i, j, startX, startY,
						(byte) (s + 1), orient, draw);
				} else {
					i = data[s][n];
				}
				n++;
			}
		}
	}

	public void getFormFromMenu(String string) {
		int[][] data;
		int ni = getNeighborIndex(), ri;
		Integer ruleIndex = getRuleIndex(ni);

		name = string;
		if (neighbors != Life2DRules.indexToNeighbor[ni])
			System.out.println("NeighborIndex: " + ni + ", Neighbors: " +
				neighbors);
		Life2DForms lifeForms = new Life2DForms(ni);
		if (string.contains("Random")) {
			Integer symmetryIndex = Life2DSoup.stringToSymmetryIndex(string);

			if (symmetryIndex != null) {
				soup.displayRandomSoup(
					symmetryIndex,
					fillPercent, stirs,
					Life2DSoup.soupSize(ncols),
					Life2DSoup.soupSize(nrows));
			}
			return;
		}
		if (ruleIndex == null) {
			System.out.println(string + " ruleIndex is null for neighborIndex " + ni);
			return;
		}
		ri = ruleIndex;
		data = lifeForms.get(ni, ri, string);
		if (data == null || data[0].length == 0) {
			System.out.println(string + " data for ruleIndex " + ri + " is null");
			return;
		}
		displayForm(data, true);
	}

	public void getLife() {
		this.stringSave = ((Life2DFrame) frame).readTextArea();
		if (this.stringSave == null || this.stringSave.equals("")) {
			showMessage("Empty string to read");
			return;
		}
		getLifeString();
	}

	public void getLifeFile(String fileName, StringBuilder fileData) {
		try (BufferedReader br = new BufferedReader(
				new FileReader(fileName))) {
			String lineData;

			while ((lineData = br.readLine()) != null)
				fileData.append(lineData).append(NL);
			String out = fileData.toString();
			if (out == null || out.equals("")) {
				showMessage("Empty file to read");
			}
		} catch (IOException e) {
			//e.printStackTrace();
			showMessage("Cannot read from " + fileName);
		}
	}

	public void getLifeFile() {
		StringBuilder fileData = new StringBuilder();
		getLifeFile(DATAFILE, fileData);
		this.stringSave = fileData.toString();
		getLifeString();
	}

	public void getLifeString() {
		getLifeString(this.stringSave, 0, 0);
	}

	public void getLifeString(String string,
			int x, int y) {
		try {
			String space = " ";
			String newline = NL;
			String pound = "#";
			int dim = 0, neighborhood, cells;
			int width = 5, height = 5;
			int offsetX = x, offsetY = y;
			int i = 0;
			String ruleString;
			String[] lineToken = string.split(newline);
			if (lineToken.length == 0)
				return;
			String lineData = lineToken[i];
			String[] wordToken = null;

			try {
				while (i < lineToken.length) {
					wordToken = lineData.split(space);
					if (wordToken.length > 1 &&
							(pound + "Dim").equals(wordToken[0]))
						break;
					if (++i == lineToken.length)
						return;
					lineData = lineToken[i];
				}
				if (wordToken != null && wordToken.length > 1)
					dim = Integer.parseInt(wordToken[1]);
				if (dim != 2) {
					showMessage(DATAFILE +
						" corrupted: dim " + dim);
				}
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: dim not an integer");
			}
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Neighbors
				neighborhood = Integer.parseInt(wordToken[1]);
				if (Math.abs(neighborhood) <= 2) {
					showMessage(DATAFILE +
						" corrupted: neighbors " + neighborhood);
					neighborhood = this.neighbors;
				}
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: neighbors not an integer");
				neighborhood = this.neighbors;
			}
			lineData = lineToken[++i];
			wordToken = lineData.split(space);
			// #Rule
			ruleString = Life2DRules.parseRule(neighborhood, wordToken[1]);
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Cells
				cells = Integer.parseInt(wordToken[1]);
				if (cells < 0) {
					showMessage(DATAFILE +
						" corrupted: cells " + cells);
				}
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: cells not an integer");
			}
			lineData = lineToken[++i];
			// #Name
			name = lineData.substring(lineData.indexOf(space) + 1);
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Width
				width = Integer.parseInt(wordToken[1]);
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: width not an integer");
			}
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Height
				height = Integer.parseInt(wordToken[1]);
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: height not an integer");
			}
			lineData = lineToken[++i];
			wordToken = lineData.split(space);
			// #P
			if (wordToken.length > 1)
				offsetX += Integer.parseInt(wordToken[1]);
			if (wordToken.length > 2)
				offsetY += Integer.parseInt(wordToken[2]);
			//((Life2DFrame) frame).callback(
			//	Life2DInterface.ACTION_RESTORE);
			cells = scanField(lineToken, i, offsetX, offsetY);
			drawField(true);
			System.out.println(DATAFILE +
				": neighbors " + neighborhood +
				", rule " + ruleString +
				", cells " + cells +
				", name " + name +
				", width " + width +
				", height " + height +
				", offsetX " + offsetX +
				", offsetY " + offsetY);
		} catch (NumberFormatException e) {
			e.printStackTrace();
			showMessage("Corrupt input found when reading");
		}
	}

	public void saveInitLife() {
		initSB.setLength(0);
		printField(initSB);
	}

	public void retrieveInitLife() {
		System.out.println(initSB);
	}

	public void writeLife() {
		StringBuilder sb = new StringBuilder();

		printField(sb);
		((Life2DFrame) frame).writeTextArea(sb.toString());
	}

	public void writeLifeFile() {
		StringBuilder sb = new StringBuilder();

		printField(sb);
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				DATAFILE, false))) {
			pr.println(sb);
			System.out.println("Saved to " + DATAFILE);
		} catch (Exception e) {
			e.printStackTrace();
			showMessage("Cannot write to " + DATAFILE);
		}
	}

	public boolean findPattern() {
		String newSignature = getGridSignature();

		if (newSignature != null) {
			if (findSignatures != null) {
				int mult = findSignatures.containsSignature(newSignature);

				if (mult != 0) {
					System.out.println("found (" + mult + ")!");
					System.out.println(newSignature);
					printLife();
					return (true);
				}
			}
			if (findNCells != null) {
				int numberCells = Signatures.signatureToNumberLiveCells(newSignature);
				for (int i = 0; i < findNCells.length; i++) {
					if (findNCells[i] == numberCells && (signatures == null ||
							signatures.containsSignature(newSignature) == 0)) {
						checkFind = i; //findPeriod[i];
						//System.out.println("found: " + i + " " + findPeriod[i]);
					}
				}
			}
		}
		return (signatures != null && signatures.containsSignature(newSignature) != 0);
	}

	public synchronized boolean printSearch(int iter) {
		if (iter == 0) {
			stamp = getGridStamp();
			signature = getGridSignature();
			oldnCells = nCells = Signatures.signatureToNumberLiveCells(signature);
			//System.out.println(signature + " in db");
			return (signatures == null || signatures.containsSignature(signature) == 0);
		}

		String newStamp = getGridStamp();
		String newSignature = getGridSignature();
		nCells = Signatures.signatureToNumberLiveCells(newSignature);
		int blob = (int) ((1.0 - (double) cCells / (2 * nCells)) * 100.0);

		if ((oldnCells <= searchCount || nCells <= searchCount) &&
				blob <= blobPercent &&
				((findStable && cCells == 0 && nCells != 0 &&
				newStamp.equals(stamp) &&
				newSignature.equals(signature)) ||
				/*!findStable &&*/
				(oldnCells != nCells ||
				!newStamp.equals(stamp) ||
				!newSignature.equals(signature))) &&
				(excludeCount < 2 ||
				oldnCells % excludeCount != 0 ||
				nCells % excludeCount != 0)) {
			if (signatures != null && signatures.containsSignature(newSignature) != 0) {
				return false;
			}
			if (stamp != null)
				System.out.println(stamp);
			if (newStamp != null)
				System.out.println(newStamp);
			if (signature != null &&
					!signature.equals(newSignature)) {
				System.out.println(signature);
				if (signatures != null)
					signatures.putSignature(signature);
			}
			if (newSignature != null) {
				System.out.println(newSignature);
				if (signatures != null)
					signatures.putSignature(newSignature);
			}
			if (genOffset > 0)
				System.out.println("Edge hit");
			printLife();
		}
		return true;
	}

	public void printLife() {
		StringBuilder sb = new StringBuilder();

		printField(sb);
		System.out.println(sb);
	}

	void eraseFrame(Graphics g) {
		g.setColor(background);
		g.fillRect(0, 0, getSize().width, getSize().height);
	}

	void eraseFrame() {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			eraseFrame(g);
		} finally {
			g.dispose();
		}
	}

	void draw3DFrame(boolean inside, int x, int y, int width, int height) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			if (inside)
				g.setColor(background.darker());
			else
				g.setColor(background);
			g.fillRect(x, y, width, 1);
			g.fillRect(x, y + 1, 1, height - 1);
			g.fillRect(x + 1, y + 1, width - 2, 1);
			g.fillRect(x + 1, y + 2, 1, height - 3);
			if (inside)
				g.setColor(background.brighter().brighter());
			else
				g.setColor(background);
			g.fillRect(x + 1, y + height - 1, width - 1, 1);
			g.fillRect(x + width - 1, y + 1, 1, height - 2);
			g.fillRect(x + 2, y + height - 2, width - 3, 1);
			g.fillRect(x + width - 2, y + 2, 1, height - 4);
		} finally {
			g.dispose();
		}
	}

	void drawFrame(Graphics g, boolean show) {
		Rectangle b = cellPl.boundRect(ncols, nrows);

		if (g == null || b == null)
			return;
		g.setColor(background.darker().darker());
		g.fillRect(0, 0, getSize().width, b.y);
		g.fillRect(0, b.y, b.x, b.height - b.y);
		g.fillRect(0, b.height, getSize().width, b.y);
		g.fillRect(b.width, b.y,
			getSize().width - b.width, b.height - b.y);
		draw3DFrame(show && focus, b.x, b.y,
			b.width - b.x, b.height - b.y);
	}

	void drawFrame(boolean show) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			drawFrame(g, show);
		} finally {
			g.dispose();
		}
	}

	void recalibrateBounds(int width, int height) {
		boolean first = true;
		minCol = width / 2 - 1;
		maxCol = width / 2 + 1;
		minRow = height / 2 - 1;
		maxRow = height / 2 + 1;

		for (CellInfo cellInfo : field.list) {
			int position = cellInfo.getPosition();
			int col = getColumn(position, width);
			int row = getRow(position, width);
			if (first) {
				minCol = maxCol = col;
				if (col > 0)
					minCol--;
				if (col < width - 1)
					maxCol++;
				minRow = maxRow = row;
				if (row > 0)
					minRow--;
				if (row < height - 1)
					maxRow++;
				first = false;
			} else {
				if (col - 1 < minCol && col > 0)
					minCol = col - 1;
				if (col + 1 > maxCol && col < width - 1)
					maxCol = col + 1;
				if (row - 1 < minRow && row > 0)
					minRow = row - 1;
				if (row + 1 > maxRow && row < height - 1)
					maxRow = row + 1;
			}
		}
	}

	boolean checkAliveBounds(int i, int j) {
		if (minCol == 0 || maxCol == ncols - 1 ||
				i == 0 || i == ncols - 1) {
			minCol = 0;
			maxCol = ncols - 1;
			edge = true;
		} else if (i <= minCol) {
			minCol = i - 1;
		} else if (i >= maxCol) {
			maxCol = i + 1;
		}
		if (minRow == 0 || maxRow == nrows - 1 ||
				j == 0 || j == nrows - 1) {
			minRow = 0;
			maxRow = nrows - 1;
			edge = true;
		} else if (j <= minRow) {
			minRow = j - 1;
		} else if (j >= maxRow) {
			maxRow = j + 1;
		}
		return edge;
	}

	int scanField(String[] lineToken, int stringIndex,
			int offsetX, int offsetY) {
		int polygon = getPolygon();
		String lineData;
		int m = stringIndex, count = 0;
		int i, j, n, p, startX, startY, row = 0;
		String space = " ";

		/*for (i = 0; i < lineToken.length; i++) {
			for (j = 0; j < lineToken[i].length(); j++) {
				if (lineToken[i].charAt(j) == comment) {
					break;
				}
				if (lineToken[i].charAt(j) == endOfPatternRLE) {
					return scanFieldRLE(lineToken, stringIndex, offsetX, offsetY);
				}
			}
		}*/
		startX = offsetX + ncols / 2;
		startY = offsetY + nrows / 2;
		if (polygon == 3) {
			if (((startX + startY) & 1) == 1) {
				startX--;
			}
		} else if (polygon == 5) {
			startX -= (startX + 2 * startY) % 4;
		}
		try {
			for (j = startY; ++m < lineToken.length; j++) {
				lineData = lineToken[m];
				if (lineData.length() == 0) {
					break; // allows flexibility
				}
				n = 0;
				i = startX;
				if (polygon == 3) {
					if (((startX + startY) & 1) == 1)
						i--;
				} else if (polygon == 6) {
					i -= (row + (startY & 1)) / 2;
				}
				while (n < lineData.length()) {
					byte oldCell, newCell;

					if (space.charAt(0) == lineData.charAt(n)) {
						n++;
						continue;
					}
					/*if (i < 0 || i >= ncols ||
							j < 0 || j >= nrows) {
						n++;
						continue;
					}*/
					p = getPosition(i, j);
					newCell = 0;
					if (states == 2) {
						newCell = (byte) ((Life2DInterface.deadChar(lineData.charAt(n))) ? 0 : 1);
					} else {
						if (!Life2DInterface.deadChar(lineData.charAt(n))) {
							newCell = (byte) (lineData.charAt(n) - '0');
						}
					}
					if (newCell != 0) {
						count++;
					}
					oldCell = field.getCell(p);
					if (oldCell != newCell) {
						field.setCell(p, newCell);
						CellInfo info = new CellInfo(
							p, newCell);
						drawCell(info);
						if (newCell == 0) {
							nCells--;
						} else if (oldCell == 0) {
							nCells++;
							checkAliveBounds(i, j);
						}
						((Life2DFrame) frame).updateCells(nCells);
					}
					i++;
					n++;
				}
				row++;
			}
		} catch (Exception e) {
			e.printStackTrace();
			showMessage("Corrupt input found when reading (scanField)");
		}
		return count;
	}

	/*int scanFieldRLE(String[] lineToken, int stringIndex,
			int offsetX, int offsetY) {
		final char blank = 'b';
		final char occupied = 'o';
		final char line = '$';
		int polygon = getPolygon();
		String lineData;
		int m = stringIndex, count = 0;
		int i, j, n, p, startX, startY, row = 0;
		int digit = 0;
		final int BASE = 10;
		String space = " ";

		startX = offsetX + ncols / 2;
		startY = offsetY + nrows / 2;
		if (polygon == 3) {
			if (((startX + startY) & 1) == 1) {
				startX--;
			}
		} else if (polygon == 5) {
			startX -= (startX + 2 * startY) % 4;
		}
		for (i = 0; i < lineToken.length; i++) {
			lineData = lineToken[i];
			for (j = 0; j < lineData.length(); j++) {
				if (lineData.charAt(j) == comment) {
					break;
				}
				if (lineData.charAt(j) == endOfPatternRLE) {
					return count;
				}
				p = getPosition(i, j);
				if (lineData.charAt(j) == blank) {
					// do something
					byte newCell = (byte) 0;
					byte oldCell = field.getCell(p);
					if (oldCell != newCell) {
						p = getPosition(i, j);
						field.setCell(p, newCell);
						CellInfo info = new CellInfo(
							p, newCell);
						drawCell(info);
						if (newCell == 0) {
							nCells--;
						} else if (oldCell == 0) {
							nCells++;
							checkAliveBounds(i, j);
						}
						((Life2DFrame) frame).updateCells(nCells);
					}
					digit = 0;
				} else if (lineData.charAt(j) == occupied) {
					// do something
					digit = 0;
				} else if (lineData.charAt(j) == endOfLine) {
					row++;
					digit = 0; // if number > 1 can use for multi-dimension
				} else if (lineData.charAt(j) >= '0' &&
						lineData.charAt(j) <= '9') {
					digit = digit * BASE + lineData.charAt(j) - '0';
				}
			}
		}
		return count;
	}*/

	public synchronized void drawField(boolean draw) {
		nCells = cellCount();
		for (CellInfo cellInfo : field.list) {
			if (draw)
				drawCell(cellInfo);
		}
		if (draw) {
			if (grid)
				drawGrid();
			((Life2DFrame) frame).updateCells(nCells);
			capturer.show();
		}
	}

	public void redrawField() {
		if (!scroll) {
			eraseFrame();
			drawFrame(true);
		}
		drawField(true);
	}

	public Rectangle getSelectedRectangle() {
		return capturer.getRectangle();
	}

	void banner(String msg) {
		((Life2DFrame) frame).displayMessage(msg);
	}

	void drawGrid() {
		Graphics g = this.getGraphics();
		Color color;

		if (g == null)
			return;
		if (grid)
			color = Color.gray;
		else
			color = background;
		g.setColor(color);
		for (int j = 0; j < nrows; j++)
		for (int i = 0; i < ncols; i++) {
				cellPl.drawCell(g, false, null, null, i, j);
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}

	@Override
	public void paint(Graphics g) {
		if (!scroll)
			drawFrame(focus);
		drawField(true);
	}

	void paintNow() {
		Graphics g = getGraphics();

		try {
			paint(g);
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			g.dispose();
		}
	}

	int getColumn(int position) {
		return position % this.ncols;
	}

	int getRow(int position) {
		return position / this.ncols;
	}

	static int getColumn(int position, int ncols) {
		return position % ncols;
	}

	static int getRow(int position, int ncols) {
		return position / ncols;
	}

	int getPosition(int col, int row) {
		return getPosition(col, row, this.ncols);
	}

	static int getPosition(int col, int row, int ncols) {
		return row * ncols + col;
	}

	int getColumn(CellInfo info) {
		return getColumn(info.getPosition());
	}

	int getRow(CellInfo info) {
		return getRow(info.getPosition());
	}

	void setPosition(CellInfo info, int col, int row) {
		info.setPosition(getPosition(col, row));
	}

	int getNeighbors(int col, int row) {
		int n, p, count = 0, gcount = 0;

		if (col == 0 || row == 0 ||
				col == ncols - 1 || row == nrows - 1) {
			edge = true;
			if (!wrap)
				return 0;
		}
		for (n = 0; n < Math.abs(neighbors); n++) {
			p = positionOfNeighbor(n, col, row);
			gcount <<= 1;
			int state = field.getCell(p);

			if (states == 2) {
				if (state != 0) {
					count++; // totalistic
					gcount++; // nontotalistic
				}
			} else { // if (states == 4) // double cell
				if (state == 1 || state == 2) {
					count++; // totalistic
					//gcount++; // nontotalistic
				} else if (state == 3) {
					count += 2;
					//gcount += 2;
				}
				gcount = 0;
			}
		}
		group = gcount;
		return count;
	}

	int positionOfNeighbor(int cell, int i, int j) {
		int polygon = getPolygon();
		int col = i, row = j;
		int dir = cell * (360 / Math.abs(neighbors));

		if (neighbors < 0) { /* -6 or -9 */
			dir = (dir >= 180) ? dir - 180 : dir + 180;
		}
		switch (polygon) {
		case 3: /* TRI */
			if (((col + row) & 1) == 0) {	/* up */
				switch (dir) {
				case 0: /* close */
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 20:
				case 30:
				case 40:
					row = (row + 1 == nrows) ? 0 : row + 1;
					col = (col == 0) ? ncols - 1 : col - 1;
					break;
				case 60: /* far */
					row = (row + 1 == nrows) ? 0 : row + 1;
					col = (col < 2) ? ncols - 2 + col : col - 2;
					break;
				case 80:
				case 90:
				case 100:
					col = (col < 2) ? ncols - 2 + col : col - 2;
					break;
				case 120: /* close */
					col = (col == 0) ? ncols - 1 : col - 1;
					break;
				case 140:
				case 150:
				case 160:
					row = (row == 0) ? nrows - 1 : row - 1;
					col = (col == 0) ? ncols - 1 : col - 1;
					break;
				case 180: /* far */
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 200:
				case 210:
				case 220:
					row = (row == 0) ? nrows - 1 : row - 1;
					col = (col + 1 == ncols) ? 0 : col + 1;
					break;
				case 240: /* close */
					col = (col + 1 == ncols) ? 0 : col + 1;
					break;
				case 260:
				case 270:
				case 280:
					col = (col + 2 >= ncols) ? col + 2 - ncols : col + 2;
					break;
				case 300: /* far */
					row = (row + 1 == nrows) ? 0 : row + 1;
					col = (col + 2 >= ncols) ? col + 2 - ncols : col + 2;
					break;
				case 320:
				case 330:
				case 340:
					row = (row + 1 == nrows) ? 0 : row + 1;
					col = (col + 1 == ncols) ? 0 : col + 1;
					break;
				default:
					System.out.println("wrong direction " + dir);
				}
			} else {	/* down */
				switch (dir) {
				case 0: /* close */
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 20:
				case 30:
				case 40:
					row = (row == 0) ? nrows - 1 : row - 1;
					col = (col + 1 == ncols) ? 0 : col + 1;
					break;
				case 60: /* far */
					row = (row == 0) ? nrows - 1 : row - 1;
					col = (col + 2 >= ncols) ? col + 2 - ncols : col + 2;
					break;
				case 80:
				case 90:
				case 100:
					col = (col + 2 >= ncols) ? col + 2 - ncols : col + 2;
					break;
				case 120: /* close */
					col = (col + 1 == ncols) ? 0 : col + 1;
					break;
				case 140:
				case 150:
				case 160:
					row = (row + 1 == nrows) ? 0 : row + 1;
					col = (col + 1 == ncols) ? 0 : col + 1;
					break;
				case 180: /* far */
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 200:
				case 210:
				case 220:
					row = (row + 1 == nrows) ? 0 : row + 1;
					col = (col == 0) ? ncols - 1 : col - 1;
					break;
				case 240: /* close */
					col = (col == 0) ? ncols - 1 : col - 1;
					break;
				case 260:
				case 270:
				case 280:
					col = (col < 2) ? ncols - 2 + col : col - 2;
					break;
				case 300: /* far */
					row = (row == 0) ? nrows - 1 : row - 1;
					col = (col < 2) ? ncols - 2 + col : col - 2;
					break;
				case 320:
				case 330:
				case 340:
					row = (row == 0) ? nrows - 1 : row - 1;
					col = (col == 0) ? ncols - 1 : col - 1;
					break;
				default:
					System.out.println("wrong direction " + dir);
				}
			}
			break;
		case 4:
		case 6:
			switch (dir) {
			case 0:
				col = (col + 1 == ncols) ? 0 : col + 1;
				break;
			case 45:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 60:
				if ((row & 1) == 0)
					col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 90:
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 120:
				if ((row & 1) != 0)
					col = (col == 0) ? ncols - 1 : col - 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 135:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 180:
				col = (col == 0) ? ncols - 1 : col - 1;
				break;
			case 225:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			case 240:
				if ((row & 1) != 0)
					col = (col == 0) ? ncols - 1 : col - 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			case 270:
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			case 300:
				if ((row & 1) == 0)
					col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			case 315:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			default:
				System.out.println("wrong direction " + dir);
			}
			break;
		case 5:
			int orient = ((row & 1) * 2 + col) % 4;
			switch (orient) {
			case 0:
				switch (dir) {
				case 0:
					col++;
					break;
				case 51: /* 7 */
				case 72: /* 5 */
					col = (col + 2 >= ncols) ? 0 : col + 2;
					break;
				case 102: /* 7 corner */
					col = (col + 3 >= ncols) ? 1 : col + 3;
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 144: /* 5 */
				case 153: /* 7 */
					col++;
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 204: /* 7 */
				case 216: /* 5 */
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 255: /* 7 */
					col = (col == 0) ? ncols - 1 : col - 1;
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 288: /* 5 */
				case 306: /* 7 */
					col = (col == 0) ? ncols - 1 : col - 1;
					break;
				default:
					System.out.println("wrong direction " + dir);
				}
				break;
			case 1:
				switch (dir) {
				case 0:
					col--;
					break;
				case 51: /* 7 */
				case 72: /* 5 */
					col = (col == 1) ? ncols - 1 : col - 2;
					break;
				case 102: /* 7 */
					col = (col == 1) ? ncols - 2 : col - 3;
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 144: /* 5 */
				case 153: /* 7 */
					col--;
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 204: /* 7 */
				case 216: /* 5 */
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 255: /* 7 */
					col = (col + 1 >= ncols) ? 0 : col + 1;
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 288: /* 5 */
				case 306: /* 7 */
					col = (col + 1 >= ncols) ? 0 : col + 1;
					break;
				default:
					System.out.println("wrong direction " + dir);
				}
				break;
			case 2:
				switch (dir) {
				case 0:
					col = (col + 1 >= ncols) ? 0 : col + 1;
					break;
				case 51: /* 7 */
				case 72: /* 5 */
					row = (row == 0) ? nrows - 1 : row - 1;
					col++;
					break;
				case 102: /* 7 */
					col = (col == 0) ? ncols - 1 : col - 1;
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 144: /* 5 */
				case 153: /* 7 */
					col = (col == 0) ? ncols - 2 : col - 2;
					break;
				case 204: /* 7 */
				case 216: /* 5 */
					col = (col == 0) ? ncols - 1 : col - 1;
					break;
				case 255: /* 7 */
					row = (row + 1 == nrows) ? 0 : row + 1;
					col = (col == 0) ? ncols - 1 : col - 1;
					break;
				case 288: /* 5 */
				case 306: /* 7 */
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				default:
					System.out.println("wrong direction " + dir);
				}
				break;
			case 3:
				switch (dir) {
				case 0:
					col--;
					break;
				case 51: /* 7 */
				case 72: /* 5 */
					col = (col == 0) ? ncols - 1 : col - 1;
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 102: /* 7 */
					col = (col + 1 >= ncols) ? 0 : col + 1;
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 144: /* 5 */
				case 153: /* 7 */
					col = (col + 2 >= ncols) ? 1 : col + 2;
					break;
				case 204: /* 7 */
				case 216: /* 5 */
					col = (col + 1 >= ncols) ? 0 : col + 1;
					break;
				case 255: /* 7 */
					col = (col + 1 >= ncols) ? 0 : col + 1;
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 288: /* 5 */
				case 306: /* 7 */
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				default:
					System.out.println("wrong direction " + dir);
				}
				break;
			default:
				System.out.println("wrong orient " + orient);
			}
			break;
		default:
		}
		return getPosition(col, row);
	}

	/* O(n) operation for a cell but only done for mouse click. */
	byte incCell(int col, int row) {
		int p = getPosition(col, row);
		byte state = field.getCell(p);
		byte nextState = (byte) ((state < states - 1) ?
			state + 1 : 0);

		if (state == 0) {
			field.setCell(p, nextState);
		} else {
			for (Iterator<CellInfo> iterator =
					field.list.iterator();
					iterator.hasNext();) {
				CellInfo cellInfo = iterator.next();
				int position = cellInfo.getPosition();
				if (position == p) {
					field.setArrayCell(position, nextState);
					if (nextState == 0) {
						iterator.remove();
					} else {
						cellInfo.setCell(nextState);
					}
					break;
				}
			}
		}
		if (state == 0) {
			nCells++;
			checkAliveBounds(col, row);
		} else if (state == states - 1) {
			nCells--;
		}
		CellInfo info = new CellInfo(p, nextState);
		if (nextState == 0) {
			eraseCell(info);
		} else {
			drawCell(info);
		}
		((Life2DFrame) frame).updateCells(nCells);
		return nextState;
	}

	int cellCount() {
		return field.list.size();
	}

	boolean setCell(int col, int row,
			byte nextState, boolean draw) {
		int p = getPosition(col, row);
		byte state = field.getCell(p);

		if (state == 0 && nextState != 0) {
			field.setCell(p, nextState);
			nCells++;
			checkAliveBounds(col, row);
		} else if (state == nextState) {
			return false;
		} else if (nextState == 0) { // state != 0
			field.setCell(p, (byte) 0);
			nCells--;
			CellInfo info = new CellInfo(p, nextState);
			eraseCell(info);
			return false;
		} else { // different sort of live cell
			field.setCell(p, nextState);
			cCells++;
		}
		if (draw) {
			CellInfo info = new CellInfo(p, nextState);

			drawCell(info);
			if ((Life2DFrame) frame != null)
				((Life2DFrame) frame).updateCells(nCells);
		}
		return true;
	}

	/*void printState(int state) {
		CellList curr;
		int i = 0;

		curr = first[state]->next;
		System.out.println("state " + state);
		while (curr != list[state]) {
			System.out.println(i + ": position " +
				curr.info.getPosition() + ",	cell " +
				curr.info.getCell() + ", temp " +
				curr.info.getTemp());
			curr = curr.next;
			i++;
		}
	}*/

	static Color colorRainbow(int state, int factor) {
		final int MAX_INTENSITY = 255;

		if (state >= 0 && state < factor) {
			return new Color(MAX_INTENSITY, state * 256 / factor, 0);
		} else if (state >= factor && state < factor * 2) {
			return new Color(511 - state * 256 / factor, MAX_INTENSITY, 0);
		} else if (state >= factor * 2 && state < factor * 3) {
			return new Color(0, MAX_INTENSITY, state * 256 / factor - 512);
		} else if (state >= factor * 3 && state < factor * 4 - 1) { /* slight defect */
			return new Color(0, 1023 - state * 256 / factor, MAX_INTENSITY);
		} else {
			return new Color(0, 0, MAX_INTENSITY);
		}
	}

	void drawCell(CellInfo info) {
		Graphics g = this.getGraphics();
		Color color;
		int col = getColumn(info);
		int row = getRow(info);
		byte cell = info.getCell();

		if (mono) {
			color = Color.white;
		} else if (states == 2) {
			color = colorRainbow(cell, 16);
		} else {
			color = colorRainbow((cell - 1) * 16, 16);
		}
		cellPl.drawCell(g, round, color,
			(grid) ? gridColor : background, col, row);
	}

	void eraseCell(CellInfo info) {
		Graphics g = this.getGraphics();
		int col = getColumn(info);
		int row = getRow(info);

		cellPl.drawCell(g, false, background,
			(grid) ? gridColor : background, col, row);
	}

	Life2DMatrix getField() {
		return field;
	}

class Capturer {
	private Point anchor = new Point(0, 0);
	private Point movable = new Point(0, 0);
	private Point start = new Point(0, 0);
	private Point stop = new Point(0, 0);
	private int completed = -1;
	private final Color selectedColor = Color.magenta;

	Point calcPoint(int x, int y) {
		Point t = new Point(x, y);

		if (t.x < 0)
			t.x = 0;
		if (t.y < 0)
			t.y = 0;
		if (t.x >= ncols)
			t.x = ncols - 1;
		if (t.y >= nrows)
			t.y = nrows - 1;
		return t;
	}

	void calcStartPoint() {
		int polygon = getPolygon();

		if (anchor.x > movable.x)
			start.x = movable.x;
		else
			start.x = anchor.x;
		if (anchor.y > movable.y)
			start.y = movable.y;
		else
			start.y = anchor.y;
		/* System.out.println("calcStartPoint: " +
			start.x + " " + start.y); */
		if (polygon == 3) {
			if (((start.x + start.y) & 1) == 1) {
				if ((start.y & 1) == 1) {
					start.y--;
				} else {
					start.x--;
				}
			}
		} else if (polygon == 5) {
			int mod = (start.x + 2 * start.y) % 4;

			if ((mod / 2) == 1) {
				if ((start.y & 1) == 1) {
					start.y--;
				} else {
					start.x -= 2 ;
				}
			}
			start.x -= (mod & 1);
		} else if (polygon == 6 && (start.y & 1) == 1) {
			start.y--;
		}
	}

	void calcEndPoint() {
		int polygon = getPolygon();

		if (anchor.x < movable.x)
			stop.x = movable.x;
		else
			stop.x = anchor.x;
		if (anchor.y < movable.y)
			stop.y = movable.y;
		else
			stop.y = anchor.y;
		if (polygon == 3) {
			if (((stop.x + stop.y) & 1) == 1) {
				if ((stop.y & 1) == 0)
					stop.y++;
				else
					stop.x++;
			}
		} else if (polygon == 5) {
			int mod = 3 - ((stop.x + 2 * stop.y) % 4);

			if ((mod / 2) == 1) {
				if ((stop.y & 1) == 0) {
					stop.x += 2;
				} else {
					stop.y++;
				}
			}
			stop.x += (mod & 1);
		} else if (polygon == 6 && (stop.y & 1) == 0) {
			stop.y++;
		}
	}

	void startCapture(int column, int row) {
		movable = anchor = calcPoint(column, row);
		calcStartPoint();
		calcEndPoint();
		completed = 0;
		drawCaptureRect((grid) ? gridColor : background);
	}

	void continueCapture(int column, int row) {
		int calculatedColumns = getColumns();
		int calculatedRows = getRows();

		movable = calcPoint(column, row);
		drawCaptureRect((grid) ? gridColor : background);
		if (column >= 0 && row >= 0 && column < calculatedColumns && row < calculatedRows) {
			calcStartPoint();
			calcEndPoint();
			drawCaptureRect(selectedColor);
			return;
		}
		calcStartPoint();
		calcEndPoint();
	}

	void stopCapture() {
		if (completed == 0)
			completed = 1;
	}

	void clearCapture() {
		drawCaptureRect((grid) ? gridColor : background);
		anchor.setLocation(0, 0);
		movable.setLocation(0, 0);
		completed = -1;
	}

	private void drawCaptureRect(Color color) {
		Graphics g = getGraphics();
		int x = start.x;
		int y = start.y;
		int width = stop.x - start.x + 1;
		int height = stop.y - start.y + 1;

		if (completed == -1)
			return;
		cellPl.drawCaptureRect(g, color,
			x, y, width, height);
	}

	void show() {
		if (!anchor.equals(movable))
			drawCaptureRect(selectedColor);
	}

	Rectangle getRectangle() {
		if (completed == 1) {
			int x = start.x;
			int y = start.y;
			int width = stop.x - start.x + 1;
			int height = stop.y - start.y + 1;

			/*System.out.println("getRectangle: " +
				x + " " + y + " " + width + " " + height); */
			return new Rectangle(x, y, width, height);
		}
		return null;
	}
}

}
