package com.sillycycle.bagleyd.life2d;

/*
 * @(#)Life2DDialog.java
 *
 * Structure from Java Sourcebook by Ed Anuff, pp 287-289
 *
 * Copyright 1993 - 2018  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.BorderLayout;
import java.awt.Dialog;
import java.awt.GridLayout;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JPanel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.sillycycle.bagleyd.util.ColumnLayout;
import com.sillycycle.bagleyd.util.Slider;

/**
 * The <code>Life2DDialog</code> class holds the configuration
 * setup for the <code>Life2DFrame</code>.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class Life2DDialog extends Dialog {
	private static final long serialVersionUID = 42L;
	static boolean active = false;
	static Life2DFrame frame = null;

	static int columnsValue = Life2DInterface.DEFAULT_COLUMNS;
	static int neighborIndexValue = Life2DInterface.DEFAULT_NEIGHBOR_INDEX;
	static boolean randomOrientValue = true;
	static boolean gridValue = true;
	static boolean wrapValue = true;
	static boolean visibleValue = true;
	static boolean monoValue = false;
	static boolean roundValue = false;
	static int statesValue = 2;

	Slider columnsChanger = new Slider("Field Size",
		columnsValue, Life2DInterface.MIN_COLUMNS,
		Life2DInterface.MAX_COLUMNS);
	Life2DBoxes lifeBoxes = new Life2DBoxes();
	JCheckBox randomOrientCheckBox = new JCheckBox("Random Orient");
	JCheckBox gridCheckBox = new JCheckBox("Grid");
	JCheckBox wrapCheckBox = new JCheckBox("Wrap");
	JCheckBox visibleCheckBox = new JCheckBox("Visible");
	JCheckBox monoCheckBox = new JCheckBox("Mono");
	JCheckBox roundCheckBox = new JCheckBox("Round");
	JCheckBox statesCheckBox = new JCheckBox("4 state");

	public Life2DDialog(Life2DFrame frame, Frame parent,
			String titleText,
			String buttonText) {
		super(parent, titleText, true);

		JButton button = new JButton(buttonText);
		JPanel buttonPanel = new JPanel();
		JPanel sliderPanel = new JPanel();
		JPanel lifeBoxesPanel = new JPanel();
		JPanel togglePanel = new JPanel();

		columnsChanger.setWidth(150);
		buttonPanel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.CENTER));
		sliderPanel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.LEFT));
		sliderPanel.add(columnsChanger, BorderLayout.PAGE_START);
		buttonPanel.add(sliderPanel, BorderLayout.PAGE_START);
		lifeBoxesPanel.add(lifeBoxes, BorderLayout.PAGE_START);
		buttonPanel.add(lifeBoxesPanel, BorderLayout.PAGE_START);
		togglePanel.setLayout(new GridLayout(3, 0));
		randomOrientCheckBox.setSelected(randomOrientValue);
		togglePanel.add(randomOrientCheckBox);
		gridCheckBox.setSelected(gridValue);
		togglePanel.add(gridCheckBox);
		wrapCheckBox.setSelected(wrapValue);
		togglePanel.add(wrapCheckBox);
		visibleCheckBox.setSelected(visibleValue);
		togglePanel.add(visibleCheckBox);
		monoCheckBox.setSelected(monoValue);
		togglePanel.add(monoCheckBox);
		roundCheckBox.setSelected(roundValue);
		togglePanel.add(roundCheckBox);
		statesCheckBox.setSelected(statesValue != 2);
		togglePanel.add(statesCheckBox);
		buttonPanel.add(togglePanel, BorderLayout.PAGE_START);

		buttonPanel.add(button, BorderLayout.PAGE_START);
		add(BorderLayout.PAGE_START, buttonPanel);
		pack();
		Life2DDialog.active = true;
		Life2DDialog.frame = frame;
		columnsChanger.addChangeListener(new ColumnsChangeListener());
		lifeBoxes.addItemListener(new LifeItemListener());
		lifeBoxes.addActionListener(new LifeActionListener());
		randomOrientCheckBox.addItemListener(new RandomOrientListener());
		gridCheckBox.addItemListener(new GridListener());
		wrapCheckBox.addItemListener(new WrapListener());
		visibleCheckBox.addItemListener(new VisibleListener());
		monoCheckBox.addItemListener(new MonoListener());
		roundCheckBox.addItemListener(new RoundListener());
		statesCheckBox.addItemListener(new StatesListener());

		button.addActionListener((ActionEvent event) -> {
			dispose();
			//else multiple Dialogs can be opened
			active = false;
		});

		addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent event) {
				dispose();
				//else multiple Dialogs can be opened
				active = false;
			}
		});
	}

	class ColumnsChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = columnsValue;

			columnsValue = columnsChanger.getValue();
			if (oldValue != columnsValue) {
				frame.endGame();
				frame.setColumns(columnsValue);
				frame.startGame();
			}
		}
	}

	class LifeActionListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent e) {
			int ni = neighborIndexValue;

			neighborIndexValue = lifeBoxes.getNeighborIndex();
//System.out.println("Life2DDialog.actionPerformed: " + ni + " " + neighborIndexValue);
			frame.setRules(ni, neighborIndexValue);
		}
	}

	class LifeItemListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			int ni = neighborIndexValue;

			neighborIndexValue = lifeBoxes.getNeighborIndex();
//System.out.println("Life2DDialog.itemStateChanged: " + ni + " " + neighborIndexValue);
			frame.setRules(ni, neighborIndexValue);
		}
	}

	class GridListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setGrid(gridCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class RandomOrientListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setRandomOrient(randomOrientCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class WrapListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setWrap(wrapCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class VisibleListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setCellVisible(visibleCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class MonoListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setMono(monoCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class RoundListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setRound(roundCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class StatesListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setStates(statesCheckBox.isSelected() ? 4 : 2);
			frame.requestFocus();
		}
	}

	public void setColumns(int columnsValue) {
		Life2DDialog.columnsValue = columnsValue;
		columnsChanger.setValue(columnsValue);
	}

	public static int getColumns() {
		return columnsValue;
	}

	public void setStates(int statesValue) {
		// can only be 2 or 4
		if (statesValue == 2)
			Life2DDialog.statesValue = statesValue;
		else
			Life2DDialog.statesValue = 4;
		statesCheckBox.setSelected(statesValue != 2);
	}

	public static int getStates() {
		return (Life2DDialog.statesValue);
	}

	public void setNeighborIndex(int neighborIndexValue) {
		Life2DDialog.neighborIndexValue = neighborIndexValue;
		lifeBoxes.setNeighborIndex(neighborIndexValue);
	}

	public void setRuleIndex(Integer ruleIndex) {
		if (ruleIndex != null) {
			lifeBoxes.setRuleIndex(ruleIndex);
		}
	}

	public static int getNeighborIndex() {
		return neighborIndexValue;
	}

	public void setBirth(int value) {
		lifeBoxes.setBirth(value);
	}

	public void setSurvival(int value) {
		lifeBoxes.setSurvival(value);
	}

	public static void setBirthGroup(int[] value) {
		Life2DBoxes.setBirthGroup(value);
	}

	public static void setSurvivalGroup(int[] value) {
		Life2DBoxes.setSurvivalGroup(value);
	}

	public void setMenu(int neighborIndex, int birth, int survival) {
		lifeBoxes.setMenu(neighborIndex, birth, survival);
	}

	public int getBirth() {
		return lifeBoxes.getBirth();
	}

	public int getSurvival() {
		return lifeBoxes.getSurvival();
	}

	public static int[] getBirthGroup() {
		return Life2DBoxes.getBirthGroup();
	}

	public static int[] getSurvivalGroup() {
		return Life2DBoxes.getSurvivalGroup();
	}

	public void updateList() {
		lifeBoxes.updateList();
	}

	public void setRandomOrient(boolean flag) {
		Life2DDialog.randomOrientValue = flag;
		randomOrientCheckBox.setSelected(flag);
	}

	public static boolean getRandomOrient() {
		return randomOrientValue;
	}

	public void setGrid(boolean flag) {
		Life2DDialog.gridValue = flag;
		gridCheckBox.setSelected(flag);
	}

	public static boolean getGrid() {
		return gridValue;
	}

	public void setWrap(boolean flag) {
		Life2DDialog.wrapValue = flag;
		wrapCheckBox.setSelected(flag);
	}

	public static boolean getWrap() {
		return wrapValue;
	}

	public void setCellVisible(boolean flag) {
		Life2DDialog.visibleValue = flag;
		visibleCheckBox.setSelected(flag);
	}

	public static boolean getCellVisible() {
		return visibleValue;
	}

	public void setMono(boolean flag) {
		Life2DDialog.monoValue = flag;
		monoCheckBox.setSelected(flag);
	}

	public static boolean getMono() {
		return monoValue;
	}

	public void setRound(boolean flag) {
		Life2DDialog.roundValue = flag;
		roundCheckBox.setSelected(flag);
	}

	public static boolean getRound() {
		return roundValue;
	}

	public static boolean getActive() {
		return Life2DDialog.active;
	}
}
