package com.sillycycle.bagleyd.life2d;

/*
 * @(#)Life2DInterface.java
 *
 * Copyright 1993 - 2019  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

/**
 * The <code>Life2DInterface</code> class contains most of the
 * constants used by the <code>Life2DFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public final class Life2DInterface {
	static final String LIFEFILE = "out.life";
	static final String XML_FILE = "life2d.xml";

	static final int EMPTY = -1;
	static final int DEFAULT_CELLSIZE = 10;
	static final int NEXT_CELLSIZE = 3;
	static final int MIN_SPEED = 1;
	static final int MAX_SPEED = 50;
	static final int DEFAULT_COLUMNS = 100;
	static final int DEFAULT_ROWS = 75;
	static final int MIN_COLUMNS = 7;
	static final int MIN_ROWS = 4;
	static final int MAX_COLUMNS = 3 * DEFAULT_COLUMNS;
	static final int MAX_ROWS = 3 * DEFAULT_ROWS;
	static final int DEFAULT_SPEED = 40;
	static final int DEFAULT_NEIGHBOR_INDEX = 6; // For n=8
	static final char DEFAULT_ALIVE_CHAR = '*';
	static final char DEFAULT_DEAD_CHAR = '.';
	static final char[] DEAD_CHARS = {'.'};
	static final char[] ALIVE_CHARS = {'o', 'O','*', 'A', 'V', 'J', 'D'};
	static final String MSG_PAUSED = "Paused";
	static final int DEFAULT_EXCLUDE_COUNT = 1; /* exclude multiples (except 1) */
	static final int DEFAULT_SEARCH_COUNT = 64; /* largest life to catch */
	static final int DEFAULT_EXCLUDE_PERIOD = 6; /* period of life to ignore */
	static final int DEFAULT_SEARCH_LENGTH = 1024; /* max time to let gel */
	static final int DEFAULT_FILL_PERCENT = 33;
	static final int DEFAULT_BLOB_PERCENT = 60;
	static final int DEFAULT_STIRS = 0;

	static final String MSG_END = "Life Over";
	static final String TITLE = "Life2D";
	public enum Icons {X16, X22, X24, X32, X48, X64}
	static final String[] ICONS = {"icons/16x16/life2d.png",
		"icons/22x22/life2d.png",
		"icons/24x24/life2d.png",
		"icons/32x32/life2d.png",
		"icons/48x48/life2d.png",
		"icons/64x64/life2d.png"};

	static boolean aliveChar(char c) {
		for (int i = 0; i < ALIVE_CHARS.length; i++) {
			if (ALIVE_CHARS[i] == c)
				return true;
		}
		return false;
	}

	static boolean deadChar(char c) {
		for (int i = 0; i < DEAD_CHARS.length; i++) {
			if (DEAD_CHARS[i] == c)
				return true;
		}
		return false;
	}
}
