package com.sillycycle.bagleyd.life2d;

/*
 * @(#)Life2DSoup.java
 *
 * Copyright 1993 - 2014  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.util.Random;

/**
 * The <code>Life2DSoup</code> class does all the graphics for
 * <code>Life2DCanvas</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

/* I found this helpful: */
/* http://en.wikipedia.org/wiki/List_of_planar_symmetry_groups */

public class Life2DSoup {
	static Life2DCanvas canvas = null;
	Random generator;

	static final String[] symmetryStrings = {
		"Full Random", // C1
		"Sym X Random", // D1
		"Sym Y Random", // D1
		/* Non Glider Patterns, except Half Antisym Random if square */
		"Sym XY Random", // D2
		"Antisym X Random", // C2
		"Antisym Y Random", // C2
		"Half Antisym Random", // Diagonal D1, Triangular C3
		"Half Sym Random", // Mirror Diagonal D2, Mirror Triangular D3
		"Full Antisym Random", // Square C4, Hexagon C6
		"Full Sym Random", // Mirror Square D4, Mirror Hexagon D6
	};

	public Life2DSoup(Life2DCanvas canvas) {
		Life2DSoup.canvas = canvas;
		generator = new Random(System.nanoTime());
	}

	static Integer stringToSymmetryIndex(String string) {
		for (int i = 0; i < symmetryStrings.length; i++) {
			if (symmetryStrings[i].equals(string)) {
				return i;
			}
		}
		return null;
	}

	static int soupSize(int s) {
		int val = s / 4;
		final int bound = 10;
		return ((val > bound) ? val : bound);
	}

	public int randomSymmetry(int polygon, boolean gliderSearch) {
		// For glider searches, too much symmetry can cause the
		// gliders to be short lived since wrap is used to find them.
		int vector;

		if (polygon == 5) {
			int symmetryLength = 7;
			vector = 3;
			if (gliderSearch) {
				return generator.nextInt(vector);
			}
			//return generator.nextInt(symmetryLength - 1) + 1;
			return generator.nextInt(symmetryLength);
			//return generator.nextInt(symmetryLength - vector) + vector;
		}
		if (polygon == 4) {
			vector = 4;
		} else {
			vector = 3;
		}
		if (gliderSearch) {
			if (polygon == 4) {
				int symmetry = generator.nextInt(vector);

				if (symmetry == 3)
					symmetry = 6; // Diagonal Symmetry
				return symmetry;
			}
			return generator.nextInt(vector);
		}
		//return generator.nextInt(symmetryStrings.length - 1) + 1;
		return generator.nextInt(symmetryStrings.length);
		/*if (polygon == 4) {
			int symmetry = generator.nextInt(symmetryStrings.length vector) + vector;
				if (symmetry == 6)
					symmetry = 3;
				return symmetry;
		}
		return generator.nextInt(symmetryStrings.length - vector) + vector;*/
	}

	int randomState() {
		return generator.nextInt(canvas.getStates() - 1) + 1;
	}

	public void displayRandomSoup(int symmetryIndex, int percentChance, int stirs,
			int fieldX, int fieldY) {
		int ncols = canvas.getColumns();
		int nrows = canvas.getRows();
		int polygon = canvas.getPolygon();
		int col, row;
		int halfcols = ncols / 2;
		int halfrows = nrows / 2;
		int vx = fieldX / 2 + 1;
		int vy = fieldY / 2 + 1;
		int even = generator.nextInt(2);
		int yEven = generator.nextInt(2);
		int state = 1;
		int percent = percentChance;

		if (stirs > 0) {
			/* make smaller as it should expand */
			vx = (fieldX / 3) | 1;
			vy = (fieldY / 3) | 1;
		}
		if (vx < 1)
			vx = 1;
		if (vy < 1)
			vy = 1;
		//System.out.println("symmetryIndex = " + symmetryIndex);
		switch (symmetryIndex) {
		case 0: /* Full Random */ /* Ex. G */
			/* Cyclic symmetry: C1 */
			for (row = halfrows - vy; row < halfrows + vy; ++row) {
				for (col = halfcols - vx; col < halfcols + vx; ++col) {
					if (generator.nextInt(100) < percent) {
						canvas.setCell(col, row, randomState());
					}
				}
			}
			break;
		case 1: /* Sym X Random */  /* Ex. A */
			/* Dihedral symmetry: D1 */
			for (row = halfrows - vy; row < halfrows + vy; ++row) {
				for (col = halfcols; col < halfcols + vx; ++col) {
					if (generator.nextInt(100) < percent) {
						state = randomState();
						if (polygon != 5) {
							canvas.setCell(col, row, state);
						}
						switch (polygon) {
						case 3:
							canvas.setCell(2 * halfcols - col, row, state);
							break;
						case 4:
							canvas.setCell(2 * halfcols - col - even, row, state);
							break;
						case 5:
							int tcol = 2 * halfcols - col;
							int mod = (col + 2 * row) % 4;

							if (mod == 2 && col == halfcols)
								continue;
							canvas.setCell(col, row, state);
							if (mod == 1)
								tcol += 2;
							else if (mod == 2 || mod == 3)
								tcol++;
							canvas.setCell(tcol, row, state);
							break;
						case 6:
							canvas.setCell(2 * halfcols - col - (((row & 1) == 1) ? 0 : 1), row, state);
							break;
						default:
						}
					}
				}
			}
			break;
		case 2: /* Sym Y Random */ /* Ex. B */
			/* Dihedral symmetry: D1 */
			for (row = halfrows; row < halfrows + vy; ++row) {
				for (col = halfcols - vx; col < halfcols + vx; ++col) {
					if (generator.nextInt(100) < percent) {
						state = randomState();
						if (polygon != 5) {
							canvas.setCell(col, row, state);
						}
						switch (polygon) {
						case 3:
							canvas.setCell(col, 2 * halfrows - row - 1, state);
							break;
						case 4:
							canvas.setCell(col, 2 * halfrows - row - even, state);
							break;
						case 5:
							int mod = (col + 2 * row) % 4;
							int tcol = col;
							int trow = 2 * halfrows - row;

							if (mod == 1 && row == halfrows)
								continue;
							canvas.setCell(col, row, state);
							if (mod == 0)
								tcol++;
							else if (mod == 1)
								tcol--;
							canvas.setCell(tcol, trow, state);
							break;
						case 6:
							canvas.setCell(col, 2 * halfrows - row, state);
							break;
						default:
						}
					}
				}
			}
			break;
		case 3: /* Sym XY Random */ /* Ex. H */
			/* Dihedral symmetry: D2 */
			for (row = halfrows; row < halfrows + vy; ++row) {
				for (col = halfcols; col < halfcols + vx; ++col) {
					if (generator.nextInt(100) < percent) {
						state = randomState();
						if (polygon != 3 && polygon != 5) {
							canvas.setCell(col, row, state);
						}
						switch (polygon) {
						case 3:
							canvas.setCell(col + even, row, state);
							canvas.setCell(col + even, 2 * halfrows - row - 1, state);
							canvas.setCell(2 * halfcols - col + even, row, state);
							canvas.setCell(2 * halfcols - col + even, 2 * halfrows - row - 1, state);
							break;
						case 4:
							canvas.setCell(col, 2 * halfrows - row - yEven, state);
							canvas.setCell(2 * halfcols - col - even, row, state);
							canvas.setCell(2 * halfcols - col - even, 2 * halfrows - row - yEven, state);
							break;
						case 5:
							int tcol = 2 * halfcols - col;
							int trow = 2 * halfrows - row;
							int mod = (col + 2 * row) % 4;

							if ((mod == 2 && col == halfcols) ||
									(mod == 1 && row == halfrows))
								continue;
							canvas.setCell(col, row, state);
							if (mod == 1)
								tcol += 2;
							else if (mod == 2 || mod == 3)
								tcol++;
							canvas.setCell(tcol, row, state);

							tcol = col;
							if (mod == 0)
								tcol++;
							else if (mod == 1)
								tcol--;
							canvas.setCell(tcol, trow, state);

							tcol = 2 * halfcols - col + 1;
							canvas.setCell(tcol, trow, state);
							break;
						case 6:
							canvas.setCell(col, 2 * halfrows - row, state);
							canvas.setCell(2 * halfcols - col - (((row & 1) == 1) ? 0 : 1) - even, row, state);
							canvas.setCell(2 * halfcols - col - (((row & 1) == 1) ? 0 : 1) - even, 2 * halfrows - row, state);
							break;
						default:
						}
					}
				}
			}
			break;
		case 4: /* Antisym X Random */ /* Ex. S */
			/* Cyclic symmetry: C2 */
			for (row = halfrows - vy; row < halfrows + vy; ++row) {
				for (col = halfcols; col < halfcols + vx; ++col) {
					if (generator.nextInt(100) < percent) {
						if (polygon != 5) {
							canvas.setCell(col, row, state);
						}
						switch (polygon) {
						case 3:
							canvas.setCell(2 * halfcols - col, 2 * halfrows - row - 1, state);
							break;
						case 4:
							canvas.setCell(2 * halfcols - col - even, 2 * halfrows - row - yEven, state);
							break;
						case 5:
							int tcol = 2 * halfcols - col - even;
							int trow = 2 * halfrows - row - even;
							int mod = (col + 2 * row) % 4;

							if (even == 0 && mod == 2 && col == halfcols)
								continue;
							canvas.setCell(col, row, state);
							if (even == 1) {
								canvas.setCell(tcol, trow, state);
							} else {
								canvas.setCell(tcol + 1, trow, state);
							}
							break;
						case 6:
							canvas.setCell(2 * halfcols - col, 2 * halfrows - row - 1, state);
							break;
						default:
						}
					}
				}
			}
			break;
		case 5: /* Antisym Y Random */ /* Ex. S */
			/* Cyclic symmetry: C2 */
			for (row = halfrows; row < halfrows + vy; ++row) {
				for (col = halfcols - vx; col < halfcols + vx; ++col) {
					if (generator.nextInt(100) < percent) {
						state = randomState();
						if (polygon != 5) {
							canvas.setCell(col, row, state);
						}
						switch (polygon) {
						case 3:
							canvas.setCell(2 * halfcols - col - 1, 2 * halfrows - row, state);
							break;
						case 4:
							canvas.setCell(2 * halfcols - col - even, 2 * halfrows - row - yEven, state);
							break;
						case 5:
							int tcol = 2 * halfcols - col - even;
							int trow = 2 * halfrows - row - even;
							int mod = (col + 2 * row) % 4;

							if (even == 0 && mod == 0 && row == halfrows)
								continue;
							canvas.setCell(col, row, state);
							if (even == 1) {
								canvas.setCell(tcol, trow, state);
							} else {
								canvas.setCell(tcol + 1, trow, state);
							}
							break;
						case 6:
							canvas.setCell(2 * halfcols - col - 1 - (((row & 1) == 1) ? 0 : 1), 2 * halfrows - row, state);
							break;
						default:
						}
					}
				}
			}
			break;
		case 6: /* Half Antisym Random */
			// Diagonal, Triangular (rotate around face of
			// triangle or rotate around every other corner
			// of triangle, rotate around corner of hexagon
			// or every other face around hexagon
			/* Dihedral symmetry: D1 and */
			/* Cyclic symmetry: C3 (triskelion) */
			switch (polygon) {
			case 3:
				int odd = (halfcols + halfrows) & 1;

				if (even == 1) {
					/* around every other corner of triangle */
					vy = (vx + vy) / 3;
					for (row = 0; row < vy; ++row) {
						for (col = 0; col < 2 * row + 1; ++col) {
							if (generator.nextInt(100) < percent) {
								state = randomState();
								canvas.setCell(halfcols + col - row, halfrows - row - 1 + odd, state);
								canvas.setCell(halfcols - (col + 1) / 2 + 2 * row + 1, halfrows + col / 2 + odd, state);
								canvas.setCell(halfcols - col / 2 - row - 1, halfrows + row - (col + 1) / 2 + odd, state);
							}
							if (generator.nextInt(100) < percent) {
								state = randomState();
								canvas.setCell(halfcols - col + row, halfrows + row + odd, state);
								canvas.setCell(halfcols + (col + 1) / 2 - 2 * row - 1, halfrows - col / 2 - 1 + odd, state);
								canvas.setCell(halfcols + col / 2 + row + 1, halfrows - row + (col + 1) / 2 - 1 + odd, state);
							}
						}
					}
				} else {
					/* around face of triangle */
					vy = (vx + vy) / 4;
					for (row = 0; row < vy; ++row) {
						for (col = 0; col < 6 * row + 5; ++col) {
							if (generator.nextInt(100) < percent) {
								state = randomState();
								canvas.setCell(halfcols + col / 2, halfrows + (col + 1) / 2 - 2 * row - 2 + odd, state);
								canvas.setCell(halfcols + (col + 1) / 2 - 3 * row - 3, halfrows + row + 2 - col / 2 - 1 + odd, state);
								canvas.setCell(halfcols - col + 3 * row + 3, halfrows + row + 1 + odd, state);
							}
						}
					}
					if (generator.nextInt(100) < percent) {
						/* center triangle */
						canvas.setCell(halfcols, halfrows + odd, randomState());
					}
				}
				break;
			case 4:
				vx = vy = (vx + vy) / 2;
				for (row = -vy; row < vy; ++row) {
					for (col = row; col < vx; ++col) {
						if (generator.nextInt(100) < percent) {
							state = randomState();
							if (even == 1) {
								canvas.setCell(col + halfcols, row + halfrows, state);
								canvas.setCell(row + halfcols, col + halfrows, state);
							} else {
								canvas.setCell(-col + halfcols, row + halfrows, state);
								canvas.setCell(-row + halfcols, col + halfrows, state);
							}
						}
					}
				}
				break;
			case 5:
				for (row = halfrows; row < halfrows + vy; ++row) {
					for (col = halfcols; col < halfcols + 2 * vy; ++col) {
						if (generator.nextInt(100) < percent) {
							int tcol = 2 * halfcols - col - 1;
							int trow = 2 * halfrows - row - 1;
							int mod = (col + 2 * row) % 4;

							state = randomState();
							canvas.setCell(col, row, state);
							canvas.setCell(tcol, trow, state);
							tcol = (row - halfrows) * 2 + halfcols;
							trow = (col - halfcols) / 2 + halfrows;
							if (mod == 1 || mod == 2)
								tcol++;
							canvas.setCell(2 * halfcols - tcol - 1, trow, state);
							canvas.setCell(tcol, 2 * halfrows - trow - 1, state);
						}
					}
				}
				break;
			case 6:
				if (even == 1) {
					/* around corner of hexagon */
					vy = 2 * (vx + vy) / 3;
					for (row = 0; row < vy; ++row) {
						for (col = 0; col < row; ++col) {
							if (generator.nextInt(100) < percent) {
								state = randomState();
								canvas.setCell(halfcols - col + (row + (halfrows + 1) % 2) / 2, halfrows + row, state);
								canvas.setCell(halfcols + (col + row + halfrows % 2) / 2 - halfrows % 2, halfrows - row + col + 1, state);
								canvas.setCell(halfcols - row + (col + halfrows % 2) / 2 + (halfrows + 1) % 2, halfrows - col + 1, state);
							}
						}
					}
					for (row = 0; row < vy; ++row) {
						for (col = 0; col < row - 1; ++col) {
							if (generator.nextInt(100) < percent) {
								state = randomState();
								canvas.setCell(halfcols + row - (col + (halfrows + 1) % 2) / 2 - halfrows % 2, halfrows + col + 1, state);
								canvas.setCell(halfcols + col - (row + (halfrows + 1) % 2) / 2 + (halfrows + 1) % 2, halfrows - row + 1, state);
								canvas.setCell(halfcols - (col + row + halfrows % 2) / 2, halfrows + row - col, state);
							}
						}
					}
				} else {
					/* around every other face of hexagon */
					vy = 2 * (vx + vy) / 3;
					for (row = 0; row < vy; ++row) {
						for (col = 0; col < row; ++col) {
							if (generator.nextInt(100) < percent) {
								state = randomState();
								canvas.setCell(halfcols - col + (row + (halfrows + 1) % 2) / 2, halfrows + row, state);
								canvas.setCell(halfcols + (col + row + (halfrows + 1) % 2) / 2, halfrows - row + col, state);
								canvas.setCell(halfcols - row + (col + (halfrows + 1) % 2) / 2, halfrows - col, state);
							}
							if (generator.nextInt(100) < percent) {
								state = randomState();
								canvas.setCell(halfcols + row - (col + halfrows % 2) / 2, halfrows + col, state);
								canvas.setCell(halfcols + col - (row + halfrows % 2) / 2, halfrows - row, state);
								canvas.setCell(halfcols - (col + row + halfrows % 2) / 2, halfrows + row - col, state);
							}
						}
					}
					if (generator.nextInt(100) < percent) {
						canvas.setCell(halfcols, halfrows, randomState());
					}
				}
				break;
			default:
			}
			break;
		case 7: /* Half Sym Random */
			// Diagonal, Triangular (rotate around face of
			// triangle or rotate around every other corner
			// of triangle, rotate around corner of hexagon
			// or every other face around hexagon
			/* Dihedral symmetry: D2 and D3 */
			switch (polygon) {
			case 3:
				int odd = (halfcols + halfrows) & 1;

				if (even == 1) {
					/* around every other corner of triangle */
					vy = (vx + vy) / 3;
					for (row = 0; row < vy; ++row) {
						for (col = 0; col < 2 * row + 1; ++col) {
							if (generator.nextInt(100) < percent) {
								state = randomState();
								canvas.setCell(halfcols + col - row, halfrows - row - 1 + odd, state);
								canvas.setCell(halfcols - (col + 1) / 2 + 2 * row + 1, halfrows - col / 2 - 1 + odd, state);
								canvas.setCell(halfcols - col / 2 - row - 1, halfrows + row - (col + 1) / 2 + odd, state);
								canvas.setCell(halfcols + col - row, halfrows + row + odd, state);
								canvas.setCell(halfcols - (col + 1) / 2 + 2 * row + 1, halfrows + col / 2 + odd, state);
								canvas.setCell(halfcols - col / 2 - row - 1, halfrows - row + (col + 1) / 2 - 1 + odd, state);
							}
						}
					}
				} else {
					/* around face of triangle */
					vy = (vx + vy) / 4;
					for (row = 0; row < vy; ++row) {
						for (col = 0; col < 3 * row + 4; ++col) {
							if (generator.nextInt(100) < percent) {
								state = randomState();
								canvas.setCell(halfcols + col / 2, halfrows + (col + 1) / 2 - 2 * row - 2 + odd, state);
								canvas.setCell(halfcols - col / 2, halfrows + (col + 1) / 2 - 2 * row - 2 + odd, state);
								canvas.setCell(halfcols + (col + 1) / 2 - 3 * row - 3, halfrows + row + 2 - col / 2 - 1 + odd, state);
								canvas.setCell(halfcols - (col + 1) / 2 + 3 * row + 3, halfrows + row + 2 - col / 2 - 1 + odd, state);
								canvas.setCell(halfcols - col + 3 * row + 3, halfrows + row + 1 + odd, state);
								canvas.setCell(halfcols + col - 3 * row - 3, halfrows + row + 1 + odd, state);
							}
						}
					}
					if (generator.nextInt(100) < percent) {
						/* center triangle */
						canvas.setCell(halfcols, halfrows + odd, randomState());
					}
				}
				break;
			case 4:
				vy = (vx + vy) / 2;
				for (row = -vy; row < 0; ++row) {
					for (col = row; col < vx; ++col) {
						if (generator.nextInt(100) < percent) {
							state = randomState();
							canvas.setCell(col + halfcols, row + halfrows, state);
							canvas.setCell(row + halfcols, col + halfrows, state);
							canvas.setCell(-col + halfcols, -row + halfrows, state);
							canvas.setCell(-row + halfcols, -col + halfrows, state);
						}
					}
				}
				break;
			case 6:
				if (even == 1) {
					/* around corner of hexagon */
					vy = 2 * (vx + vy) / 3;
					for (row = 0; row < vy; ++row) {
						for (col = 0; col < row / 2 + 1; ++col) { // 2 6 10
							if (generator.nextInt(100) < percent) {
								state = randomState();
								canvas.setCell(halfcols + (col + row + (halfrows + 1) % 2) / 2 - 1 + halfrows % 2, halfrows - row + col + 2, state);
								canvas.setCell(halfcols + row - (col + (halfrows + 1) % 2) / 2 - 1, halfrows - col + 1, state);
								canvas.setCell(halfcols - col + (row + halfrows % 2) / 2, halfrows + row - 1, state);
								canvas.setCell(halfcols + col - (row + (halfrows + 1) % 2) / 2 + 1, halfrows + row - 1, state);
								canvas.setCell(halfcols - row + (col + halfrows % 2) / 2 + 2, halfrows - col + 1, state);
								canvas.setCell(halfcols - (col + row + halfrows % 2) / 2 + 1 + halfrows % 2, halfrows - row + col + 2, state);
							}
						}
					}
					for (row = 0; row < vy; ++row) {
						for (col = 0; col < row / 2; ++col) {
							if (generator.nextInt(100) < percent) { // 0 4 8
								state = randomState();
								canvas.setCell(halfcols + col - (row + halfrows % 2) / 2 + 1 + halfrows % 2, halfrows - row + 2, state);
								canvas.setCell(halfcols - col + (row + (halfrows + 1) % 2) / 2 + halfrows % 2 - 1, halfrows - row + 2, state);
								canvas.setCell(halfcols + row - (col + (halfrows + 1) % 2) / 2 - 1, halfrows + col + 1, state);
								canvas.setCell(halfcols + (col + row + halfrows % 2) / 2, halfrows + row - col - 1, state);
								canvas.setCell(halfcols - (col + row + (halfrows + 1) % 2) / 2 + 1, halfrows + row - col - 1, state);
								canvas.setCell(halfcols - row + (col + halfrows % 2) / 2 + 2, halfrows + col + 1, state);
							}
						}
					}
				} else {
					/* around every other face of hexagon */
					vy = 2 * (vx + vy) / 3;
					for (row = 0; row < vy; ++row) {
						for (col = 0; col <= row; ++col) {
							if (generator.nextInt(100) < percent) {
								state = randomState();
								canvas.setCell(halfcols + col - (row + halfrows % 2) / 2, halfrows + row, state);
								canvas.setCell(halfcols + row - (col + halfrows % 2) / 2, halfrows - col, state);
								canvas.setCell(halfcols - (col + row + halfrows % 2) / 2, halfrows - row + col, state);
								canvas.setCell(halfcols + row - (col + halfrows % 2) / 2, halfrows + col, state);
								canvas.setCell(halfcols + col - (row + halfrows % 2) / 2, halfrows - row, state);
								canvas.setCell(halfcols - (col + row + halfrows % 2) / 2, halfrows + row - col, state);
							}
						}
					}
					if (generator.nextInt(100) < percent) {
						canvas.setCell(halfcols, halfrows, randomState());
					}
				}
				break;
			default:
			}
			break;
		case 8: /* Full Antisym Random */
			/* Cyclic Symmetry: C4 (swastika), C6 */
			// Square, Hexagonal (rotate around corner of
			// triangle, rotate around face of hexagon)
			switch (polygon) {
			case 3:
				int odd = (halfcols + halfrows) & 1;

				vy = (vx + vy) / 3;
				for (row = 0; row < vy; ++row) {
					for (col = 0; col < 2 * row + 1; ++col) {
						if (generator.nextInt(100) < percent) {
							state = randomState();
							canvas.setCell(halfcols + col / 2 + row + 1, halfrows - row + (col + 1) / 2 - 1 + odd, state); /* 1 */
							canvas.setCell(halfcols - (col + 1) / 2 + 2 * row + 1, halfrows + col / 2 + odd, state); /* 3 */
							canvas.setCell(halfcols - col + row, halfrows + row + odd, state); /* 5 */
							canvas.setCell(halfcols - col / 2 - row - 1, halfrows + row - (col + 1) / 2 + odd, state); /* 7 */
							canvas.setCell(halfcols + (col + 1) / 2 - 2 * row - 1, halfrows - col / 2 - 1 + odd, state); /* 9 */
							canvas.setCell(halfcols + col - row, halfrows - row - 1 + odd, state); /* 11 */
						}
					}
				}
				break;
			case 4:
				vy = (vx + vy) / 2;
				for (row = -vy; row < 0; ++row) {
					for (col = row; col < vx; ++col) {
						if (generator.nextInt(100) < percent) {
							state = randomState();
							canvas.setCell(col + halfcols + even, row + halfrows + even, state);
							canvas.setCell(row + halfcols + even, -col + halfrows, state);
							canvas.setCell(-col + halfcols, -row + halfrows, state);
							canvas.setCell(-row + halfcols, col + halfrows + even, state);
						}
					}
				}
				if (even == 0 && generator.nextInt(100) < percent) {
					canvas.setCell(halfcols, halfrows, randomState());
				}
				break;
			case 6:
				vy = 2 * (vx + vy) / 3;
				for (row = 0; row < vy; ++row) {
					for (col = 0; col < row; ++col) {
						if (generator.nextInt(100) < percent) {
							state = randomState();
							canvas.setCell(halfcols - col + (row + ((halfrows + 1) & 1)) / 2, halfrows + row, state);
							canvas.setCell(halfcols + (col + row + ((halfrows + 1) & 1)) / 2, halfrows - row + col, state);
							canvas.setCell(halfcols - row + (col + ((halfrows + 1) & 1)) / 2, halfrows - col, state);
							canvas.setCell(halfcols + row - (col + halfrows % 2) / 2, halfrows + col, state);
							canvas.setCell(halfcols + col - (row + halfrows % 2) / 2, halfrows - row, state);
							canvas.setCell(halfcols - (col + row + halfrows % 2) / 2, halfrows + row - col, state);
						}
					}
				}
				if (generator.nextInt(100) < percent) {
					canvas.setCell(halfcols, halfrows, randomState());
				}
				break;
			default:
			}
			break;
		case 9: /* Full Sym Random */
			/* Dihedral symmetry: D4 and D6 */
			// Square, Hexagonal (rotate around corner of
			// triangle, rotate around face of hexagon)
			switch (polygon) {
			case 3:
				int odd = (halfcols + halfrows) & 1;

				vy = (vx + vy) / 3;
				for (row = 0; row < vy; ++row) {
					for (col = 0; col < row + 1; ++col) {
						if (generator.nextInt(100) < percent) {
							state = randomState();
							canvas.setCell(halfcols - col + row, halfrows - row - 1 + odd, state); /* 0 */
							canvas.setCell(halfcols + col / 2 + row + 1, halfrows - row + (col + 1) / 2 - 1 + odd, state); /* 1 */
							canvas.setCell(halfcols + (col + 1) / 2 - 2 * row - 1, halfrows + col / 2 + odd, state); /* 2 */
							canvas.setCell(halfcols - (col + 1) / 2 + 2 * row + 1, halfrows + col / 2 + odd, state); /* 3 */
							canvas.setCell(halfcols - col / 2 - row - 1, halfrows - row + (col + 1) / 2 - 1 + odd, state); /* 4 */
							canvas.setCell(halfcols - col + row, halfrows + row + odd, state); /* 5 */
							canvas.setCell(halfcols + col - row, halfrows + row + odd, state); /* 6 */
							canvas.setCell(halfcols - col / 2 - row - 1, halfrows + row - (col + 1) / 2 + odd, state); /* 7 */
							canvas.setCell(halfcols - (col + 1) / 2 + 2 * row + 1, halfrows - col / 2 - 1 + odd, state); /* 8 */
							canvas.setCell(halfcols + (col + 1) / 2 - 2 * row - 1, halfrows - col / 2 - 1 + odd, state); /* 9 */
							canvas.setCell(halfcols + col / 2 + row + 1, halfrows + row - (col + 1) / 2 + odd, state); /* 10 */
							canvas.setCell(halfcols + col - row, halfrows - row - 1 + odd, state); /* 11 */
						}
					}
				}
				break;
			case 4:
				vy = (vx + vy) / 2;
				for (row = -vy; row < 0; ++row) {
					for (col = row; col <= 0; ++col) {
						if (generator.nextInt(100) < percent) {
							state = randomState();
							canvas.setCell(col + halfcols + even, row + halfrows + even, state);
							canvas.setCell(row + halfcols + even, -col + halfrows, state);
							canvas.setCell(-col + halfcols, -row + halfrows, state);
							canvas.setCell(-row + halfcols, col + halfrows + even, state);
							canvas.setCell(-col + halfcols, row + halfrows + even, state);
							canvas.setCell(-row + halfcols, -col + halfrows, state);
							canvas.setCell(col + halfcols + even, -row + halfrows, state);
							canvas.setCell(row + halfcols + even, col + halfrows + even, state);
						}
					}
				}
				if (even == 0 && generator.nextInt(100) < percent) {
					canvas.setCell(halfcols, halfrows, randomState());
				}
				break;
			case 6:
				vy = 2 * (vx + vy) / 3;
				for (row = 0; row < vy; ++row) {
					for (col = 0; col <= row / 2; ++col) {
						if (generator.nextInt(100) < percent) {
							state = randomState();
							canvas.setCell(halfcols - col + (row + (halfrows + 1) % 2) / 2, halfrows + row, state);
							canvas.setCell(halfcols + (col + row + (halfrows + 1) % 2) / 2, halfrows - row + col, state);
							canvas.setCell(halfcols - row + (col + (halfrows + 1) % 2) / 2, halfrows - col, state);
							canvas.setCell(halfcols + row - (col + halfrows % 2) / 2, halfrows + col, state);
							canvas.setCell(halfcols + col - (row + halfrows % 2) / 2, halfrows - row, state);
							canvas.setCell(halfcols - (col + row + halfrows % 2) / 2, halfrows + row - col, state);
							canvas.setCell(halfcols + col - (row + halfrows % 2) / 2, halfrows + row, state);
							canvas.setCell(halfcols + row - (col + halfrows % 2) / 2, halfrows - col, state);
							canvas.setCell(halfcols - (col + row + halfrows % 2) / 2, halfrows - row + col, state);
							canvas.setCell(halfcols - row + (col + (halfrows + 1) % 2) / 2, halfrows + col, state);
							canvas.setCell(halfcols + (col + row + (halfrows + 1) % 2) / 2, halfrows + row - col, state);
							canvas.setCell(halfcols - col + (row + (halfrows + 1) % 2) / 2, halfrows - row, state);
						}
					}
				}
				if (generator.nextInt(100) < percent) {
					canvas.setCell(halfcols, halfrows, randomState());
				}
				break;
			default:
			}
			break;
		default:
			System.out.println("wrong symmetry index " +
				symmetryIndex);
		}
	}
}
