package com.sillycycle.bagleyd.util;

/*
 * @(#)CellPn.java
 *
 * Copyright 2009 - 2016  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;

/**
 * The <code>CellPn</code> class holds the colors and
 * handles the construction of the pentagon cells (Cairo Tiling).
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class CellPn extends Component {
	private static final long serialVersionUID = 42L;

	static final int MAX_ORIENT = 4;
	static final int MAX_SIDES = 5;
	static final int MAX_EYES = 2;

	int[][] pentagonListX = new int[MAX_ORIENT][MAX_SIDES];
	int[][] pentagonListY = new int[MAX_ORIENT][MAX_SIDES];

	Rectangle r;

	CellPn(int nx, int ny, int screenWidth, int screenHeight, boolean scroll) {
		int offsetX, offsetY;
		int startX, startY;
		int xs, ys;

		if (scroll) {
			offsetX = 16;
			offsetY = 16;
			xs = (screenWidth - offsetX) / nx;
			ys = (screenHeight - offsetY) / ny;
			startX = 0;
			startY = 0;
		} else {
			offsetX = 16;
			offsetY = 6;
			xs = (screenWidth - offsetX) / nx;
			ys = (screenHeight - offsetY) / ny;
			startX = (screenWidth - offsetX - xs * nx) / 2;
			startY = (screenHeight - offsetY - ys * ny) / 2;
		}
		if (xs < 1)
			xs = 1;
		if (ys < 1)
			ys = 1;
		r = new Rectangle(startX, startY, xs, ys);

		Point[][] pentagonUnit = new Point[MAX_ORIENT][MAX_SIDES];

		/* down pt */
		pentagonUnit[0][0] = new Point(0, 0);
		pentagonUnit[0][1] = new Point(-2, -1);
		pentagonUnit[0][2] = new Point(1, -2);
		pentagonUnit[0][3] = new Point(2, 0);
		pentagonUnit[0][4] = new Point(1, 2);
		/* left pt */
		pentagonUnit[1][0] = new Point(0, 0);
		pentagonUnit[1][1] = new Point(1, -2);
		pentagonUnit[1][2] = new Point(2, 1);
		pentagonUnit[1][3] = new Point(0, 2);
		pentagonUnit[1][4] = new Point(-2, 1);
		/* up pt */
		pentagonUnit[2][0] = new Point(0, 0);
		pentagonUnit[2][1] = new Point(2, 1);
		pentagonUnit[2][2] = new Point(-1, 2);
		pentagonUnit[2][3] = new Point(-2, 0);
		pentagonUnit[2][4] = new Point(-1, -2);
		/* right pt */
		pentagonUnit[3][0] = new Point(0, 0);
		pentagonUnit[3][1] = new Point(-1, 2);
		pentagonUnit[3][2] = new Point(-2, -1);
		pentagonUnit[3][3] = new Point(0, -2);
		pentagonUnit[3][4] = new Point(2, -1);

		for (int l = 0; l < MAX_ORIENT; l++)
			for (int k = 0; k < MAX_SIDES; k++) {
				pentagonListX[l][k] = r.width *
					pentagonUnit[l][k].x / 2 +
					((k > 0) ? pentagonListX[l][k - 1] : 0);
				pentagonListY[l][k] = r.height *
					pentagonUnit[l][k].y / 4 +
					((k > 0) ? pentagonListY[l][k - 1] : 0);
		}
	}

	Rectangle boundRect(int nx, int ny) {
		return new Rectangle(r.x - 3, r.y - 2,
			r.x + r.width * nx + 13,
			r.y + r.height * ny + 8);
	}

	Point screenToCell(int x, int y) {
		int xo = x - 5;
		int yo = y - 2;
		int mod = (xo - r.x) / (2 * r.width);
		Point p = new Point();

		p.y = ModArith.div(yo - r.y, r.height);
		if (((mod + p.y) & 1) == 1) {
			p.x = (xo - r.x) / r.width;
		} else {
			p.x = 2 * mod + (((yo - r.y) %
				r.height >= r.height / 2) ? 1 : 0);
		}
		return p;
	}

	void drawCell(Graphics g, boolean round, Color cid, Color bid,
			int i, int j) {
		int startX, startY, orient; /* down left up right */
		int[] map = {2, 0, 1, 3};
		int offsetX = 0, offsetY = 0;
		int offsetXR = 0, offsetYR = 0, offsetRX = 0, offsetRY = 0;

		if (g == null)
			return;
		orient = ((j & 1) * 2 + i) % 4;
		switch (orient) {
		case 0:
			offsetX = r.width;
			offsetRY = r.height / 4;
			offsetXR = -r.width / 2;
			break;
		case 1:
			offsetY = 3 * r.height / 2;
			offsetRY = r.height / 4;
			offsetXR = -r.width / 2;
			offsetYR = -r.height + offsetRY;
			break;
		case 2:
			offsetX = -r.width / 2;
			offsetY = 3 * r.height / 4;
			offsetRX = r.width / 4;
			offsetXR = offsetRX;
			offsetYR = -r.height / 2;
			break;
		case 3:
			offsetX = 3 * r.width / 2;
			offsetY = 3 * r.height / 4;
			offsetRX = r.width / 4;
			offsetXR = -r.width;
			offsetYR = -r.height / 2;
			break;
		default:
			System.out.println("wrong orient " + orient);
		}
		orient = map[orient];
		startX = offsetX + r.x + r.width * i + 5;
		startY = offsetY + r.y + r.height * j;
		if (cid != null) {
			g.setColor(cid);
			if (round) {
				 g.fillOval(startX + offsetXR,
					startY + offsetYR,
					r.width - offsetRX,
					r.height - offsetRY);
			} else {
				fillPolygon(g, startX, startY,
					pentagonListX[orient], pentagonListY[orient],
					MAX_SIDES);
			}
		}
		if (bid != null || cid == null) {
			if (bid != null)
				g.setColor(bid);
			drawPolygon(g, startX, startY,
				pentagonListX[orient], pentagonListY[orient],
				MAX_SIDES);
		}
	}

/*		switch (orient) {
		case 0:
			offsetX = r.width;
			break;
		case 1:
			offsetY = r.height;
			break;
		case 2:
			offsetY = r.height / 2;
			break;
		case 3:
			offsetX = r.width;
			offsetY = r.height / 2;
			break;
		default:
			System.out.println("wrong orient " + orient);
		}*/

	static void fillPolygon(Graphics g, int startX, int startY,
			int[] listX, int[] listY, int n) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		g.fillPolygon(tempListX, tempListY, n);
	}

	static void drawPolygon(Graphics g, int startX, int startY,
			int[] listX, int[] listY, int n) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];
		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		g.drawPolygon(tempListX, tempListY, n);
	}

	void drawCaptureRect(Graphics g, Color color,
			int i, int j, int nx, int ny) {
		if (g == null)
			return;
		g.setColor(color);
		/* Not sure how to draw crooked rectangle.
		 * Choosing to draw each pentagon instead. */
		/*g.drawRect(r.x + nx * i, y + ny * j,
			r.width * nx, r.height * ny);*/
		for (int vert = j; vert < ny + j; vert++) {
			for (int horiz = i; horiz < nx + i; horiz++) {
				drawCell(g, false, null, null, horiz, vert);
			}
		}
	}

	void drawEyes(Graphics g, Color color, short direction,
			int i, int j) {
		int startX, startY, orient; /* down left up right */
		int[] map = {2, 0, 1, 3};
		int offsetX = 0, offsetY = 0;

		int ang = direction * MAX_SIDES / 360;
		orient = ((j & 1) * 2 + i) % 4;
		switch (orient) {
		case 0:
			offsetX = r.width;
			break;
		case 1:
			offsetY = 3 * r.height / 2;
			break;
		case 2:
			offsetX = -r.width / 2;
			offsetY = 3 * r.height / 4;
			break;
		case 3:
			offsetX = 3 * r.width / 2;
			offsetY = 3 * r.height / 4;
			break;
		default:
			System.out.println("wrong orient " + orient);
		}
		orient = map[orient];
		startX = offsetX + r.x + r.width * i + 5;
		startY = offsetY + r.y + r.height * j;

		g.setColor(color);
		Point center = new Point(0, 0);
		for (int side = 0; side < MAX_SIDES; side++) {
			if (side != 0) {
				center.x += pentagonListX[orient][side];
				center.y += pentagonListY[orient][side];
			}
		}
		center.x = center.x / MAX_SIDES + startX;
		center.y = center.y / MAX_SIDES + startY;
		for (int eye = 0; eye < MAX_EYES; eye++) {
			int side = (MAX_SIDES - ang + 2 * eye + 3) % MAX_SIDES;
			Point poly = new Point(startX, startY);

			if (side != 0) {
				poly.x += pentagonListX[orient][side];
				poly.y += pentagonListY[orient][side];
			}
			g.fillRect((poly.x + center.x) / 2,
				(poly.y + center.y) / 2, 2, 2);
		}
	}

	void drawTruchet(Graphics g, Color color, int truchetState,
			boolean sharpTurn, int i, int j) {
		int startX, startY, orient; /* down left up right */
		int[] map = {2, 0, 1, 3};
		int offsetX = 0, offsetY = 0;

		if (g == null)
			return;
		orient = ((j & 1) * 2 + i) % 4;
		switch (orient) {
		case 0:
			offsetX = r.width;
			break;
		case 1:
			offsetY = 3 * r.height / 2;
			break;
		case 2:
			offsetX = -r.width / 2;
			offsetY = 3 * r.height / 4;
			break;
		case 3:
			offsetX = 3 * r.width / 2;
			offsetY = 3 * r.height / 4;
			break;
		default:
			System.out.println("wrong orient " + orient);
		}
		orient = map[orient];
		startX = offsetX + r.x + r.width * i + 5;
		startY = offsetY + r.y + r.height * j;

		int side;

		g.setColor(color);
		int skip = (sharpTurn) ? 1 : 2;
		Point[] poly = new Point[MAX_SIDES];
		Point center = new Point();
		for (side = 0; side < MAX_SIDES; side++) {
			poly[side] = new Point();
			poly[side].x = startX + (pentagonListX[orient][side]
				+ pentagonListX[orient][(side + MAX_SIDES - 1) % MAX_SIDES]) / 2;
			poly[side].y = startY + (pentagonListY[orient][side]
				+ pentagonListY[orient][(side + MAX_SIDES - 1) % MAX_SIDES]) / 2;
			center.x += poly[side].x;
			center.y += poly[side].y;
		}
		for (side = 0; side < MAX_SIDES; side++) {
			poly[side] = new Point();
			poly[side].x = startX + (pentagonListX[orient][side]
				+ pentagonListX[orient][(side + MAX_SIDES - 1) % MAX_SIDES]) / 2;
			poly[side].y = startY + (pentagonListY[orient][side]
				+ pentagonListY[orient][(side + MAX_SIDES - 1) % MAX_SIDES]) / 2;
		}
		center.x /= MAX_SIDES;
		center.y /= MAX_SIDES;
		Point[] poly2 = new Point[MAX_SIDES];
		for (side = 0; side < MAX_SIDES; side++) {
			poly2[side] = new Point();
			poly2[side].x = (poly[side].x + skip * center.x) / (skip + 1);
			poly2[side].y = (poly[side].y + skip * center.y) / (skip + 1);
		}
		side = truchetState;
		int side2 = (side + skip) % MAX_SIDES;
		g.drawLine(poly[side].x, poly[side].y,
			poly2[side].x, poly2[side].y);
		g.drawLine(poly2[side].x, poly2[side].y,
			poly2[side2].x, poly2[side2].y);
		g.drawLine(poly2[side2].x, poly2[side2].y,
			poly[side2].x, poly[side2].y);
		/*g.setColor(bid);
		drawPolygon(g, startX, startY,
			pentagonListX[orient], pentagonListY[orient],
			MAX_SIDES);*/ // fix drawing errors

	}
}
